##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""IObjectMover implementation.
"""
from __future__ import absolute_import, print_function

import io
import os
import struct

from zope.interface import implementer


from ..interfaces import IObjectMover
from ..mover import AbstractObjectMover
from ..mover import RowBatcherStoreTemps
from ..mover import metricmethod_sampled

class PostgreSQLRowBatcherStoreTemps(RowBatcherStoreTemps):
    generic_suffix = """
    ON CONFLICT (zoid) DO UPDATE
    SET state = excluded.state,
        prev_tid = excluded.prev_tid,
        md5 = excluded.md5
    """


@implementer(IObjectMover)
class PostgreSQLObjectMover(AbstractObjectMover):

    def __init__(self, *args, **kwargs):
        super(PostgreSQLObjectMover, self).__init__(*args, **kwargs)
        if not self.driver.supports_copy:
            batcher = PostgreSQLRowBatcherStoreTemps(self.keep_history,
                                                     self.driver.Binary,
                                                     self.make_batcher)
            self.replace_temps = batcher.replace_temps
            self.store_temps = batcher.store_temps

    @metricmethod_sampled
    def on_store_opened(self, cursor, restart=False):
        """Create the temporary tables for storing objects"""
        # Note that the md5 column is not used if self.keep_history == False.
        # Ideally we wouldn't execute any of these on a restart, but
        # I've seen an issue with temp_store apparently going missing on pg8000.
        #
        # In testing, using 'DELETE ROWS' is faster than using 'DROP TABLE':
        # 1230/840ms for zodbshootout add/update with DROP vs 983/681ms.
        # The theory was that maybe a vacuum or analyze would be needed after
        # DELETE ROWS and not DROP TABLE, but that didn't seem to be true (it's possible
        # an ANALYZE would still be helpful before using the temp table, but we
        # haven't benchmarked that).
        if not restart:
            temp_store_table_tmpl = """
            CREATE TEMPORARY TABLE IF NOT EXISTS {NAME} (
                zoid        BIGINT NOT NULL PRIMARY KEY,
                prev_tid    BIGINT NOT NULL,
                md5         CHAR(32),
                state       BYTEA
            ) ON COMMIT DELETE ROWS;
            """
            ddl_stmts = [
                temp_store_table_tmpl.format(NAME='temp_store'),
                temp_store_table_tmpl.format(NAME='temp_store_replacements'),
                """
                CREATE TEMPORARY TABLE IF NOT EXISTS temp_blob_chunk (
                    zoid        BIGINT NOT NULL,
                    chunk_num   BIGINT NOT NULL,
                    chunk       OID,
                    PRIMARY KEY (zoid, chunk_num)
                ) ON COMMIT DELETE ROWS;
                """,
                """
                -- This trigger removes blobs that get replaced before being
                -- moved to blob_chunk.  Note that it is never called when
                -- the temp_blob_chunk table is being dropped or truncated.
                CREATE TRIGGER temp_blob_chunk_delete
                    BEFORE DELETE ON temp_blob_chunk
                    FOR EACH ROW
                    EXECUTE PROCEDURE temp_blob_chunk_delete_trigger();
                """,
            ]

            for stmt in ddl_stmts:
                cursor.execute(stmt)
            cursor.connection.commit()

        AbstractObjectMover.on_store_opened(self, cursor, restart)

    @metricmethod_sampled
    def restore(self, cursor, batcher, oid, tid, data):
        """Store an object directly, without conflict detection.

        Used for copying transactions into this database.
        """
        if self.keep_history:
            suffix = """
            ON CONFLICT (zoid, tid) DO UPDATE SET
                tid = excluded.tid,
                prev_tid = excluded.prev_tid,
                md5 = excluded.md5,
                state_size = excluded.state_size,
                state = excluded.state
            """
        else:
            suffix = """
            ON CONFLICT (zoid) DO UPDATE SET
                tid = excluded.tid,
                state_size = excluded.state_size,
                state = excluded.state
            """
        self._generic_restore(batcher, oid, tid, data,
                              command='INSERT', suffix=suffix)

    @metricmethod_sampled
    def download_blob(self, cursor, oid, tid, filename):
        """Download a blob into a file."""
        stmt = """
        SELECT chunk
        FROM blob_chunk
        WHERE zoid = %s
            AND tid = %s
        ORDER BY chunk_num
        """
        # Beginning in RelStorage 3, we no longer chunk blobs.
        # All chunks were collapsed into one as part of the migration.
        bytecount = 0
        cursor.execute(stmt, (oid, tid))
        rows = cursor.fetchall()
        assert len(rows) == 1
        loid, = rows[0]

        blob = cursor.connection.lobject(loid, 'rb')
        # Use the native psycopg2 blob export functionality
        blob.export(filename)
        blob.close()
        bytecount = os.path.getsize(filename)
        return bytecount

    @metricmethod_sampled
    def upload_blob(self, cursor, oid, tid, filename):
        """Upload a blob from a file.

        If serial is None, upload to the temporary table.
        """
        # pylint:disable=too-many-branches,too-many-locals
        if tid is not None:
            if self.keep_history:
                delete_stmt = """
                DELETE FROM blob_chunk
                WHERE zoid = %s AND tid = %s
                """
                cursor.execute(delete_stmt, (oid, tid))
            else:
                delete_stmt = "DELETE FROM blob_chunk WHERE zoid = %s"
                cursor.execute(delete_stmt, (oid,))

            use_tid = True
            insert_stmt = """
            INSERT INTO blob_chunk (zoid, tid, chunk_num, chunk)
            VALUES (%(oid)s, %(tid)s, %(chunk_num)s, %(loid)s)
            """

        else:
            use_tid = False
            delete_stmt = "DELETE FROM temp_blob_chunk WHERE zoid = %s"
            cursor.execute(delete_stmt, (oid,))

            insert_stmt = """
            INSERT INTO temp_blob_chunk (zoid, chunk_num, chunk)
            VALUES (%(oid)s, %(chunk_num)s, %(loid)s)
            """

        # Since we only run on 9.6 and above, the sizes of large objects
        # are allowed to exceed 2GB (int_32). The server is already chunking
        # large objects internally by itself into 4KB pages, so there's no
        # advantage to us also adding a layer of chunking.
        #
        # As long as we keep our usage simple, that's fine. Only
        # blob.seek(), blob.truncate() and blob.tell() have a need to
        # use a specific 64-bit function. `export()` and `import()`
        # (called implicitly by creating the lobject with a local
        # filename in psycopg2) work with small fixed buffers (8KB) and
        # don't care about filesize or offset; they just need the
        # `open` and `read` syscalls to handle 64-bit files (and don't
        # they have to for Python to handle 64-bit files?)
        #
        # psycopg2 explicitly uses the 64 family of functions;
        # psycopg2cffi does *not* but if it's built on 64-bit
        # platform, that's fine. pg8000 uses the SQL interfaces, not
        # the libpq interfaces, and that's also fine. Since we don't use
        # any of the functions that need 64-bit aware, none of that should be an
        # issue.

        # Create and upload the blob, getting a large object identifier.
        blob = cursor.connection.lobject(0, 'wb', 0, filename)
        blob.close()

        # Now put it into our blob_chunk table.
        params = dict(oid=oid, chunk_num=0, loid=blob.oid)
        if use_tid:
            params['tid'] = tid
        cursor.execute(insert_stmt, params)


    def _do_store_temps(self, cursor, state_oid_tid_iter, table_name):
        # History-preserving storages need the md5 to compare states.
        # We could calculate that on the server using pgcrypto, if its
        # available. Or we could just compare directly, instead of comparing
        # md5; that's fast on PostgreSQL.
        if state_oid_tid_iter:
            buf = TempStoreCopyBuffer(table_name,
                                      state_oid_tid_iter,
                                      self._compute_md5sum if self.keep_history else None)
            cursor.copy_expert(buf.COPY_COMMAND, buf)

    @metricmethod_sampled
    def store_temps(self, cursor, state_oid_tid_iter):
        self._do_store_temps(cursor, state_oid_tid_iter, 'temp_store')

    @metricmethod_sampled
    def replace_temps(self, cursor, state_oid_tid_iter):
        # Upload and then replace. We *could* go right into the table
        # if we first deleted but that would require either iterating twice
        # and/or bufferring all the state data in memory. If it's small that's ok,
        # but it could be large.
        self._do_store_temps(cursor, state_oid_tid_iter, 'temp_store_replacements')
        # TODO: Prepare this query.
        cursor.execute(
            """
            UPDATE temp_store
            SET prev_tid = r.prev_tid,
                md5 = r.md5,
                state = r.state
            FROM temp_store_replacements r
            WHERE temp_store.zoid = r.zoid
            """
        )


class TempStoreCopyBuffer(io.BufferedIOBase):
    """
    A binary file-like object for putting data into
    ``temp_store``.
    """

    # pg8000 uses readinto(); psycopg2 uses read().

    COPY_COMMAND_TMPL = "COPY {NAME} (zoid, prev_tid, md5, state) FROM STDIN WITH (FORMAT binary)"

    def __init__(self, table, state_oid_tid_iterable, digester):
        super(TempStoreCopyBuffer, self).__init__()
        self.COPY_COMMAND = self.COPY_COMMAND_TMPL.format(NAME=table)
        self.state_oid_tid_iterable = state_oid_tid_iterable
        self._iter = iter(state_oid_tid_iterable)
        self._digester = digester
        if digester and bytes is not str:
            # On Python 3, this outputs a str, but our protocol needs bytes
            self._digester = lambda s: digester(s).encode("ascii")
        if self._digester:
            self._read_tuple = self._read_one_tuple_md5
        else:
            self._read_tuple = self._read_one_tuple_no_md5

        self._done = False
        self._header = self.HEADER
        self._buffer = bytearray(8192)


    SIGNATURE = b'PGCOPY\n\xff\r\n\0'
    FLAGS = struct.pack("!i", 0)
    EXTENSION_LEN = struct.pack("!i", 0)
    HEADER = SIGNATURE + FLAGS + EXTENSION_LEN
    # All tuples begin with their length in 16 signed bits, which is the same for all tuples
    # (zoid, prev_tid, md5, state)
    _common = "!hiqiqi"
    WITH_SUM = struct.Struct(_common + "32si")
    NO_SUM = struct.Struct(_common + "i")
    # Each column in the tuple is a 32-bit length (-1
    # for NULL), followed by exactly that many bytes of data.
    # Each column datum is written in binary format; for character
    # fields (like md5) that turns out to be a direct dump of the ascii.
    # For BIGINT fields, that's an 8-byte big-endian encoding
    # For BYTEA fields, it's just the raw data
    # Finally, the trailer is a tuple size of -1
    TRAILER = struct.pack("!h", -1)

    def read(self, size=-1):
        # We don't handle "read everything in one go".
        # assert size is not None and size > 0
        if self._done:
            return b''

        if len(self._buffer) < size:
            self._buffer.extend(bytearray(size - len(self._buffer)))

        count = self.readinto(self._buffer)
        if not count:
            return b''
        return bytes(self._buffer)

    def readinto(self, buf):
        # We basically ignore the size of the buffer,
        # writing more into it if we need to.
        if self._done:
            return 0

        requested = len(buf)
        # bytearray.clear() is only in Python 3
        del buf[:]

        buf.extend(self._header)
        self._header = b''

        while len(buf) < requested:
            try:
                self._read_tuple(buf)
            except StopIteration:
                buf.extend(self.TRAILER)
                self._done = True
                break

        return len(buf)

    def __len__(self):
        return len(self.state_oid_tid_iterable)

    def _read_one_tuple_md5(self,
                            buf,
                            _pack_into=WITH_SUM.pack_into,
                            _header_size=WITH_SUM.size,
                            _blank_header=bytearray(WITH_SUM.size)):

        data, oid_int, tid_int = next(self._iter)
        len_data = len(data)
        md5 = self._digester(data)
        offset = len(buf)
        buf.extend(_blank_header)
        _pack_into(
            buf, offset,
            4,
            8, oid_int,
            8, tid_int,
            32, md5,
            len_data
        )
        buf.extend(data)

    def _read_one_tuple_no_md5(self,
                               buf,
                               _pack_into=NO_SUM.pack_into,
                               _header_size=NO_SUM.size,
                               _blank_header=bytearray(NO_SUM.size)):
        data, oid_int, tid_int = next(self._iter)
        len_data = len(data)
        offset = len(buf)
        buf.extend(_blank_header)
        _pack_into(
            buf, offset,
            4,
            8, oid_int,
            8, tid_int,
            -1,
            len_data
        )
        buf.extend(data)



##############################################################################
#
# Copyright (c) 2016 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import print_function

import os

import doctest
import re
import unittest
import shutil

import ZODB.tests.util
import zope.testing.renormalizing
import zope.testing.setupstack

import random2

from relstorage.cache.storage_cache import StorageCache

from relstorage.cache.tests import MockOptions, MockAdapter
from relstorage.storage.tpc.temporary_storage import TemporaryStorage

def _build_history():
    random = random2.Random(42)
    history = []
    serial = 2
    for i in range(1000):
        serial += 1
        oid = random.randint(i + 1000, i + 6000)
        history.append((b's', oid, serial,
                        b'x' * random.randint(200, 2000)))
        for _ in range(10):
            oid = random.randint(i + 1000, i + 6000)
            history.append((b'l', oid, serial,
                            b'x' * random.randint(200, 2000)))
    return history

def _send_queue(self, tid_int, temp_objects):
    self.cache.set_all_for_tid(tid_int, temp_objects)
    temp_objects.reset()

now = 0

def cache_run(name, size):
    history = _build_history()
    serial = 1
    #random.seed(42)
    global now
    now = 1278864701.5
    options = MockOptions()
    options.cache_local_mb = size * (1<<20)
    options.cache_local_dir = '.'
    options.cache_local_compression = 'zlib'
    # We can interleave between instances
    adapter = MockAdapter()
    poller = adapter.poller
    cache = StorageCache(adapter, options, name)
    new_cache = cache.new_instance()
    assert hasattr(cache.cache, 'tracer'), cache.cache
    assert hasattr(new_cache.cache, 'tracer'), new_cache.cache
    assert new_cache.cache.tracer is cache.cache.tracer
    assert new_cache.adapter is cache.adapter
    assert new_cache.adapter.poller is cache.adapter.poller

    all_oids = {row[1] for row in history}

    try:
        # Establish polling
        cache.poll(None, None, None)
        # poll again, having determined all of these oids have been created.
        poller.poll_tid += 1
        poller.poll_changes = [
            (oid, poller.poll_tid) for oid in sorted(all_oids)
        ]
        cache.poll(None, None, None)
        # Poll again for no changes
        poller.poll_changes = []
        new_cache.poll(None, None, None)

        new_cache_ts = TemporaryStorage()
        cache_ts = TemporaryStorage()
        data_tot = 0
        for action, oid, serial, data in history:
            data_tot += len(data)
            now += 1
            if action == b's':
                cache_ts.store_temp(oid, data)
                #print("Storing", oid, serial)
                _send_queue(cache, serial, cache_ts)
                cache.adapter.mover.data[oid] = (data, serial)
                poller.poll_tid = serial
                poller.poll_changes = [(oid, serial)]
                cache.poll(None, None, None)
                poller.poll_changes = []
                new_cache.poll(None, None, None)
                assert cache.highest_visible_tid == serial
            else:
                v = new_cache.load(None, oid)
                if v[0] is None:
                    #print("Store after miss", oid, 1)
                    new_cache_ts.store_temp(oid, data)
                    _send_queue(new_cache, 1, new_cache_ts)
                    cache.adapter.mover.data[oid] = (data, 1)
    finally:
        cache.close()

def setUp(test):
    import time
    import ZEO.scripts.cache_stats
    import ZEO.scripts.cache_simul
    # setupstack doesn't ignore problems when files can't be
    # found
    # XXX: Restore this.
    zope.testing.setupstack.rmtree = lambda p: shutil.rmtree(p, True)

    zope.testing.setupstack.setUpDirectory(test)
    def ctime(t):
        return time.asctime(time.gmtime(t-3600*4))

    assert ZEO.scripts.cache_stats.ctime is time.ctime
    ZEO.scripts.cache_stats.ctime = ctime
    assert ZEO.scripts.cache_simul.ctime is time.ctime
    ZEO.scripts.cache_simul.ctime = ctime

    os.environ["ZEO_CACHE_TRACE"] = 'single'
    test.globs['timetime'] = time.time
    time.time = lambda: now

    #cache_run('cache', 2)

def tearDown(test):
    import time
    import ZEO.scripts.cache_stats
    import ZEO.scripts.cache_simul

    ZEO.scripts.cache_stats.ctime = time.ctime
    ZEO.scripts.cache_simul.ctime = time.ctime
    os.environ.pop('ZEO_CACHE_TRACE')

    time.time = test.globs['timetime']
    zope.testing.setupstack.tearDown(test)

def test_suite():
    suite = unittest.TestSuite()
    try:
        __import__('ZEO')
    except ImportError:
        class NoTest(unittest.TestCase):
            @unittest.skip("ZEO not installed")
            def test_cache_trace_analysis(self):
                "Does nothing"
        suite.addTest(unittest.makeSuite(NoTest))
    else:
        suite.addTest(
            doctest.DocFileSuite(
                'cache_trace_analysis.rst',
                setUp=setUp,
                tearDown=tearDown,
                checker=ZODB.tests.util.checker + \
                    zope.testing.renormalizing.RENormalizing([
                        (re.compile(r'31\.3%'), '31.2%'),
                    ]),
                )
            )

    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""The core of RelStorage, a ZODB storage for relational databases.

Stores pickles in the database.
"""
from __future__ import absolute_import
from __future__ import print_function

import logging
import sys
import weakref


import ZODB.interfaces


from ZODB import ConflictResolution

from ZODB.mvccadapter import HistoricalStorageAdapter

from ZODB.POSException import ReadConflictError
from ZODB.POSException import ReadOnlyError
from ZODB.POSException import ReadOnlyHistoryError

from ZODB.utils import z64
from zope import interface
from zope.interface import implementer

from ..blobhelper import BlobHelper
from ..blobhelper.interfaces import IBlobHelper
from ..blobhelper.interfaces import INoBlobHelper

from ..cache.storage_cache import StorageCache
from ..cache.interfaces import CacheConsistencyError
from ..options import Options
from ..interfaces import IRelStorage
from ..adapters.connections import LoadConnection
from ..adapters.connections import StoreConnectionPool
from ..adapters.connections import ClosedConnection
from ..adapters.connections import ClosedConnectionPool
from ..adapters.connections import SingleConnectionPool
from .._compat import clear_frames
from .._compat import metricmethod
from .._util import int64_to_8bytes
from .._util import bytes8_to_int64
from .._compat import OID_SET_TYPE

from .transaction_iterator import HistoryFreeTransactionIterator
from .transaction_iterator import HistoryPreservingTransactionIterator

from .copy import Copy
from .history import History
from .history import UndoableHistory
from .legacy import LegacyMethodsMixin
from .load import Loader
from .load import BlobLoader
from .oid import OIDs
from .oid import ReadOnlyOIDs
from .pack import Pack
from .store import Storer
from .store import BlobStorer

from .tpc import NotInTransaction
from .tpc.begin import HistoryFree
from .tpc.begin import HistoryPreserving
from .tpc.restore import Restore

from .util import copy_storage_methods
from .util import make_cannot_write
from .interfaces import StorageDisconnectedDuringCommit

__all__ = [
    'RelStorage',
]

log = logger = logging.getLogger("relstorage")


class _ClosedCache(object):
    __slots__ = ()

    def close(self):
        "does nothing"

    release = close
    object_index = ()
    highest_visible_tid = None
    stats = lambda s: {'closed': True}
    afterCompletion = lambda s, c: None


@implementer(IRelStorage)
class RelStorage(LegacyMethodsMixin,
                 ConflictResolution.ConflictResolvingStorage):

    """
    Storage to a relational database, based on invalidation polling.
    """

    # pylint:disable=too-many-instance-attributes,too-many-public-methods

    _adapter = None
    _options = None
    _is_read_only = False
    _read_only_error = ReadOnlyError
    # ZODB TID of the last transaction committed by this instance.
    _last_tid_i_committed_bytes = z64

    # _closed is True after self.close() is called.
    _closed = False

    # _cache if set is a StorageCache object.
    _cache = None

    # If the blob directory is set, blobhelper is a BlobHelper.
    blobhelper = BlobHelper(None, None)

    # The state of committing that were in. Certain operations are
    # only available in certain states; certain information is only needed
    # in certain states.
    _tpc_phase = None

    __name__ = None

    # _instances is a list of weak references to storage instances bound
    # to the same database.
    _instances = ()

    _load_connection = ClosedConnection()
    _store_connection_pool = ClosedConnectionPool()

    _oids = ReadOnlyOIDs()

    def __init__(self, adapter, name=None, create=None,
                 options=None, cache=None, blobhelper=None,
                 store_connection_pool=None,
                 **kwoptions):
        # pylint:disable=too-many-branches, too-many-statements
        if options and kwoptions:
            raise TypeError("The RelStorage constructor accepts either "
                            "an options parameter or keyword arguments, not both")

        self._adapter = adapter

        if options is None:
            options = Options(**kwoptions)
        self._options = options

        if not name:
            name = options.name
            if not name:
                name = 'RelStorage: %s' % adapter
        self.__name__ = name

        self._is_read_only = options.read_only

        need_check_compat = create is None
        if create is None:
            create = options.create_schema

        if create:
            self._adapter.schema.prepare()
        elif need_check_compat:
            # At the top level, not new_instance(), and not asked to create.
            self._adapter.schema.verify()


        # A ZODB Connection is documented as not being thread-safe and
        # must be used only by a single thread at a time. In IMVCC,
        # each Connection has a unique storage object. So the storage
        # object is used only by a single thread. So there's really
        # not much point in RelStorage keeping thread locks around its
        # methods.

        # There are a small handful of ZODB tests that assume the
        # storage is thread safe (but they know nothing about IMVCC);
        # for those tests we use a custom wrapper storage that makes
        # a single instance appear thread safe.

        self._instances = []

        self._load_connection = LoadConnection(self._adapter.connmanager)
        self._load_connection.on_first_use = self.__on_load_first_use
        self.__queued_changes = OID_SET_TYPE()
        if store_connection_pool is not None:
            self._store_connection_pool = store_connection_pool
        elif not self._is_read_only:
            self._store_connection_pool = StoreConnectionPool(self._adapter.connmanager)

        if cache is not None:
            self._cache = cache
        else:
            prefix = options.cache_prefix
            if not prefix:
                # Use the database name as the cache prefix.
                with self._load_connection.isolated_connection() as cur:
                    prefix = adapter.schema.get_database_name(cur)

                prefix = prefix.replace(' ', '_')
                options.cache_prefix = prefix
            self._cache = StorageCache(adapter, options, prefix)

        if blobhelper is not None:
            self.blobhelper = blobhelper
        else:
            self.blobhelper = BlobHelper(options=options, adapter=adapter)

        tpc_begin_factory = HistoryPreserving if self._options.keep_history else HistoryFree

        if hasattr(self._adapter.packundo, 'deleteObject'):
            interface.alsoProvides(self, ZODB.interfaces.IExternalGC)

        self._tpc_phase = NotInTransaction(tpc_begin_factory, self._is_read_only)
        if not self._is_read_only:
            self._oids = OIDs(self._adapter.oidallocator)

        # Now copy in a bunch of methods from our component objects.
        # Many of these are 'stale_aware', meaning that we can ask
        # them for a version of themselves that does something
        # different when we go stale. When that happens, we *replace*
        # the object in our __dict__ with the new one; and then
        # reverse that when we're no longer stale.
        #
        # The storage wrapper zc.zlibstorage also copies methods into
        # itself when it is created, from the storage it is wrapping.
        # Because of this, stale aware methods like history() do not
        # do the right thing when we're wrapped by zc.zlibstorage.
        loader = Loader(self._adapter, self._load_connection, self._cache)
        copy_storage_methods(self, loader)
        storer = Storer()
        copy_storage_methods(self, storer)

        if options.keep_history:
            interface.alsoProvides(self, ZODB.interfaces.IStorageUndoable)
            history = UndoableHistory(self._adapter, self._load_connection)
        else:
            history = History(self._adapter, self._load_connection)
        copy_storage_methods(self, history)

        assert IBlobHelper.providedBy(self.blobhelper)
        if not INoBlobHelper.providedBy(self.blobhelper):
            interface.alsoProvides(self, ZODB.interfaces.IBlobStorageRestoreable)

            loader = BlobLoader(self._load_connection, self.blobhelper)
            copy_storage_methods(self, loader)

            storer = BlobStorer()
            copy_storage_methods(self, storer)

    @property
    def keep_history(self):
        return self._options.keep_history

    def __repr__(self):
        return "<%s at %x keep_history=%s phase=%r blobhelper=%r cache=%r>" % (
            self.__class__.__name__,
            id(self),
            self.keep_history,
            self._tpc_phase,
            self.blobhelper,
            self._cache
        )

    def new_instance(self, before=None):
        """Creates and returns another storage instance.

        See ZODB.interfaces.IMVCCStorage.
        """
        options = self._options
        if before and not self._options.read_only:
            options = self._options.new_instance(read_only=True)
        adapter = self._adapter.new_instance()
        cache = self._cache.new_instance(before=before, adapter=adapter)
        blobhelper = self.blobhelper.new_instance(adapter=adapter)
        other = type(self)(adapter=adapter, name=self.__name__,
                           create=False, options=options, cache=cache,
                           blobhelper=blobhelper,
                           store_connection_pool=self._store_connection_pool.new_instance())
        if before:
            other._read_only_error = ReadOnlyHistoryError
            other.tpc_begin = make_cannot_write(other, other.tpc_begin)
        # NOTE: We're depending on the GIL (or list implementation)
        # for thread safety here.
        self._instances.append(weakref.ref(other, self._instances.remove))

        if self._crs_transform_record_data is not type(self)._crs_transform_record_data:
            # registerDB has been called on us but isn't called on
            # our children. Make sure any wrapper that needs to transform
            # records can do so.
            # See https://github.com/zodb/relstorage/issues/71
            other._crs_transform_record_data = self._crs_transform_record_data
            other._crs_untransform_record_data = self._crs_untransform_record_data
        return other

    def before_instance(self, before):
        """
        Return a historical connection for the given *before* tid.

        The connection is read-only, and that is enforced at this level.
        """
        # XXX This might need some work to better adapt the cache? We
        # can *know* that we're pinned to see older transactions, so
        # our TID doesn't need to move forward, and we don't need to
        # poll. We're read-only so we'll never be joined to a
        # transaction or commit, we'll never have invalidations. The
        # HistoricalStorageAdapter blocks sync() and
        # poll_invalidations() from being called on this instance, but
        # our load connection is going to do that automatically the
        # first time it's accessed after a transaction
        # rollback...which, since we're not joined to a transaction,
        # should never happen...except for afterCompletion(), which
        # the ZODB connection has called on it for every transaction
        # whether it's joined or not...but the HistoricalStorageAdapter
        # doesn't implement that *either*, so it never makes it down to
        # this level. This means our load connection can stay open and
        # un-rolled-back for quite a long time, which is probably not great.
        i = self.new_instance(before=before)
        x = HistoricalStorageAdapter(i, before)
        return x

    @property
    def highest_visible_tid(self):
        cache_tid = self._cache.highest_visible_tid or 0
        committed_tid = bytes8_to_int64(self._last_tid_i_committed_bytes)
        # In case we haven't polled yet.
        return max(cache_tid, committed_tid)

    def zap_all(self, **kwargs):
        """Clear all objects and transactions out of the database.

        Used by the test suite and the ZODBConvert script.
        """
        self._load_connection.drop()
        self._store_connection_pool.drop_all()
        self._adapter.schema.zap_all(**kwargs)
        self._cache.zap_all()

    def release(self):
        """
        Release external resources used by this storage instance.

        This includes the database sessions (connections) and any memcache
        connections.

        Does *not* affect any other instances created by this instance. This object
        should still be :meth:`close` (but note that might have global affects
        on other instances of the same base object).
        """
        self._load_connection.drop()
        self._store_connection_pool.release()

        self._cache.release()
        self._cache = _ClosedCache()
        self._tpc_phase.close()
        self._tpc_phase = _ClosedCache()
        self._oids = None
        self._load_connection = ClosedConnection()
        self._store_connection_pool = ClosedConnectionPool()
        self._adapter.release()
        if not self._instances:
            self._closed = True

    def close(self):
        """Close the storage and all instances."""
        if self._closed:
            return

        self._closed = True
        self._load_connection.drop()
        self._store_connection_pool.close()
        self._load_connection = ClosedConnection()
        self._store_connection_pool = ClosedConnectionPool()

        self.blobhelper.close()
        for wref in self._instances:
            instance = wref()
            if instance is not None:
                instance.close()
        self._instances = ()
        logger.debug("Closing storage cache with stats %s", self._cache.stats())
        self._cache.close()
        self._cache = _ClosedCache()
        self._tpc_phase.close()
        self._tpc_phase = _ClosedCache()
        self._oids = None
        self._adapter.close()

    def __len__(self):
        return self._adapter.stats.get_object_count()

    def sortKey(self):
        """Return a string that can be used to sort storage instances.

        The key must uniquely identify a storage and must be the same
        across multiple instantiations of the same storage.
        """
        return self.__name__

    def getName(self):
        return self.__name__

    def getSize(self):
        """Return database size in bytes"""
        return self._adapter.stats.get_db_size()

    def registerDB(self, wrapper):
        if (ZODB.interfaces.IStorageWrapper.providedBy(wrapper)
                # Prior to ZODB 5, this would be called by the database itself.
                # (I wish it would still do that.)
                and not ZODB.interfaces.IDatabase.providedBy(wrapper)
                and not hasattr(type(wrapper), 'new_instance')):
            # Fixes for https://github.com/zopefoundation/zc.zlibstorage/issues/2
            # We special-case zlibstorage for speed
            if hasattr(wrapper, 'base') and hasattr(wrapper, 'copied_methods'):
                type(wrapper).new_instance = _zlibstorage_new_instance
                type(wrapper).pack = _zlibstorage_pack
                from zc.zlibstorage import _Iterator
                _Iterator.__len__ = _zlibstorage_Iterator_len
                # zc.zlibstorage has a custom copyTransactionsFrom that hides
                # our own implementation. It just uses ZODb.blob.copyTransactionsFromTo.
                # Use our implementation.
                wrapper.copyTransactionsFrom = self.copyTransactionsFrom
            else:
                wrapper.new_instance = lambda s: type(wrapper)(self.new_instance())

        # Prior to ZODB 4.3.1, ConflictResolvingStorage would raise an AttributeError
        super(RelStorage, self).registerDB(wrapper)

    def isReadOnly(self):
        return self._is_read_only

    def checkCurrentSerialInTransaction(self, oid, serial, transaction):
        self._tpc_phase.checkCurrentSerialInTransaction(oid, serial, transaction)

    # For the time between tpc_begin and tpc_abort, if anything we do
    # would trigger an exception to propagate up (store(), tpc_vote(),
    # whatever) we should immediately abort the transaction at the
    # database level and release locks. This will allow other
    # transactions to make progress faster. It also can help in the ZODB test suite,
    # since some places manually call the tpc_* methods and aren't careful to
    # tpc_abort() in the event of an exception like `transaction` is.
    #
    # Some of this is manual (in the tpc_* methods), but most of it is
    # done through the @phase_dependent_aborts_early decorator.

    @metricmethod
    def tpc_begin(self, transaction, tid=None, status=' '):
        try:
            self._tpc_phase = self._tpc_phase.tpc_begin(self, transaction)
        except:
            # Could be a database (connection) error, could be a programming
            # bug. Either way, we're fine to roll everything back and hope
            # for the best on a retry. Perhaps we need to raise a TransientError?
            self._load_connection.drop()
            raise

        if tid is not None:
            # tid is a committed transaction we will restore.
            # The allowed actions are carefully prescribed.
            # This argument is specified by IStorageRestoreable
            try:
                next_phase = Restore(self._tpc_phase, tid, status)
            except:
                self.tpc_abort(transaction, _force=True)
                raise
            else:
                self._tpc_phase = next_phase

    @metricmethod
    def tpc_vote(self, transaction):
        # Returns an iterable of OIDs; the storage will ghost all
        # cached objects in that list. This is invalidation because
        # the object has changed during the commit process, due to
        # conflict resolution or undo.
        try:
            next_phase = self._tpc_phase.tpc_vote(self, transaction)
        except:
            self.tpc_abort(transaction, _force=True)
            raise
        else:
            self._tpc_phase = next_phase
            return next_phase.invalidated_oids

    @metricmethod
    def tpc_finish(self, transaction, f=None):
        try:
            next_phase = self._tpc_phase.tpc_finish(self, transaction, f)
        except:
            # OH NO! This isn't supposed to happen!
            # It's unlikely tpc_abort will get called...
            self.tpc_abort(transaction, _force=True)
            raise
        # The store connection is either committed or rolledback;
        # the load connection is now rolledback.
        self._tpc_phase = next_phase
        # It might be nice to de-dup this so we're not storing it both
        # in the phase and in self, but if it was needed during TPC,
        # when our phase is not ``ITPCStateNotInTransaction``, we couldn't
        # get it.
        committed_tid = int64_to_8bytes(next_phase.last_committed_tid_int)
        self._last_tid_i_committed_bytes = committed_tid
        return committed_tid

    @metricmethod
    def tpc_abort(self, transaction, _force=False):
        # _force is not a public argument, it is an internal
        # implementation detail.
        try:
            self._tpc_phase = self._tpc_phase.tpc_abort(transaction, _force)
        except BaseException:
            if _force:
                # We're here under unexpected circumstances. It's possible something
                # might go wrong rolling back.
                self._tpc_phase = self._tpc_phase.initial_state
            raise

    def lastTransaction(self):
        if self._last_tid_i_committed_bytes == z64 and self._cache.highest_visible_tid is None:
            # We haven't committed *or* polled for transactions,
            # so our MVCC state is "floating".
            # Read directly from the database to get the latest value,
            return int64_to_8bytes(self._adapter.txncontrol.get_tid(self._load_connection.cursor))

        return max(self._last_tid_i_committed_bytes,
                   int64_to_8bytes(self._cache.highest_visible_tid or 0))

    def lastTransactionInt(self):
        return bytes8_to_int64(self.lastTransaction())

    def new_oid(self):
        # This is called from ``Connection.add`` which can be called at any time
        # by the application, so we don't know what state we're in. It is also called from
        # ``Connection._commit``, which is called during TPC.
        # If we're committing, we can't restart the connection.
        pool = self._store_connection_pool
        commit_in_progress = False
        if self._tpc_phase:
            commit_in_progress = True
            pool = SingleConnectionPool(self._tpc_phase.shared_state.store_connection)

        return self._oids.new_oid(pool, commit_in_progress)

    def iterator(self, start=None, stop=None):
        # XXX: This is broken for purposes of copyTransactionsFrom() because
        # it can only be iterated over once. zodbconvert works around this.
        if self.keep_history:
            return HistoryPreservingTransactionIterator(self._adapter, start, stop)
        return HistoryFreeTransactionIterator(
            self._adapter, self._load_connection, start, stop)

    __next = next

    def __record_iternext_gen(self, start_oid_int):
        with self._load_connection.server_side_cursor() as ss_cursor:
            for record in self._adapter.dbiter.iter_current_records(ss_cursor, start_oid_int):
                yield record

    def record_iternext(self, next=None):
        """
        Implementation of `ZODB.interfaces.IStorageCurrentRecordIteration`.

        .. caution::

           You must completely consume the iteration.
        """
        # The interface doesn't define any semantics for *next*; it's
        # completely up to the storage and is just a magic cookie.
        # However, zodbupdate assumes that the initial value for
        # `next` is interpreted like FileStorage does: as an OID (8
        # bytes) and uses that to specify an object id to start
        # with. (inclusive)
        #
        # We need to detect the case that we're given an OID; that should only happen
        # in the first call.
        start_oid_int = 0
        if isinstance(next, bytes):
            import warnings
            warnings.warn(
                "There is no defined value for the *next* parameter. "
                "RelStorage will rely on implementation-defined behaviour and treat it like "
                "FileStorage does, while assuming that iteration is beginning from the start. "
                "This may change in the future."
            )
            start_oid_int = bytes8_to_int64(next)
            next = None

        if next is None:
            # Beginning.
            cursor = self.__record_iternext_gen(start_oid_int)
            # There *should* always be at least one object: the root.
            # But if they passed a non-0 ``next`` value, that's not guaranteed.
            # However, the protocol is not very good at expressing that possibility:
            # The typical call looks like:
            #    oid, tid, data, next = storage.record_iternext(next)
            # Unpacking will raise if we return None; most processing of *data* is not expecting
            # None (it's common to wrap it in ``io.BytesIO()`` immediately; that's what zodbupdate
            # does). Thus, it's probably probably best to let getting the first value simply raise.
            oid, tid, state = self.__next(cursor)
            # After that, we can treat it like it was given to us in *next*
            next = oid, tid, state, cursor


        # Somewhere in the middle, possibly the end.
        #
        # We have to operate one ahead (keep one buffered) because
        # we need to be able to tell the caller not to call back.
        oid, tid, state, cursor = next

        try:
            new_oid, new_tid, new_state = self.__next(cursor)
        except StopIteration:
            # Signal not to call back.
            new_next = None
        else:
            new_next = new_oid, new_tid, new_state, cursor

        return oid, tid, state, new_next


    def afterCompletion(self):
        # Note that this method exists mainly to deal with read-only
        # transactions that don't go through 2-phase commit (although
        # it's called for all transactions). For this reason, we only
        # have to roll back the load connection. (The store connection
        # is completed during normal write-transaction commit or
        # abort.)

        # The next time we use the load connection, it will need to poll
        # and will call our __on_first_use.
        # Typically our next call from the ZODB Connection will be from its
        # `newTransaction` method, a forced `sync` followed by `poll_invalidations`.

        # As with `poll_invalidations`, the cache needs to know about transaction
        # boundaries like this to optimize polling for changes and invalidations.
        # If we were a write transaction, that's already done by the call to ``after_tpc_finish``,
        # but for a read-only transaction, it wouldn't happen. So for this reason,
        # as with `poll_invalidations`, the cache is in charge.

        # This doesn't use restart_and_call() or even just restart() because we don't
        # need to do those checks yet, we just want to quietly rollback.
        # They both rollback; the difference is that restart_load checks for replicas,
        # and calls any hooks needed.
        self._load_connection.rollback_quietly()

    def sync(self, force=True):
        """
        Updates to a current view of the database.

        This is implemented by rolling back the relational database
        transaction.

        .. versionchanged:: 3.0a6
           This method now pays attention to the *force* argument. If it is
           ``False``, this method does nothing. This prevents extra rollbacks
           and network traffic in the common case of using the default implicit
           transactions when this object is used by a ZODB ``Connection``.

        """
        if not force:
            # When we get a force=False, it's from Connection.afterCompletion()
            # calling Connection.newTransaction() because the transaction manager is
            # in implicit mode. Immediately after it does that, it will call
            # poll_invalidations()...and in our implementation, that implies a sync.
            # So we can avoid the overhead of the extra rollback.
            return

        rolled_back = self._load_connection.rollback_quietly()
        if not rolled_back:
            # Disconnected. Well, the rollback happens automatically
            # in that case. No big deal, ignore it.
            #
            # However, if we happened to be in the middle of committing and were
            # asked to sync (XXX: why would we?) then that's probably a problem that
            # needs to be handled and the commit rolled back too.
            if self._tpc_phase:
                raise StorageDisconnectedDuringCommit()

    def poll_invalidations(self):
        """
        Look for OIDs of objects that changed since _prev_polled_tid.

        Returns {oid: 1}, or None if all objects need to be invalidated
        because prev_polled_tid is not in the database (presumably it
        has been packed).
        """
        # poll_invalidations is called by Connection.newTransaction(), which is a
        # transaction synchronizer method.
        # Connection.afterCompletion() also calls Connection.newTransaction() if we're not
        # using explicit transactions.
        # Our sync() will have been called just before this, so we should not need to
        # restart the connection, just make a call. The load_connection logic handles that
        # for us.

        if self._closed:
            # If we've never loaded (self._load_connection is false),
            # it might seem that there's no point in doing a poll, but
            # that's not true. testMVCC breaks if we skip this. It's not exactly clear
            # why.
            return {}

        # __on_load_first_use is also the function we gave the LoadConnection to automatically
        # call when we access the cursor for the first time. In this situation, we optimize
        # for when the function is the same and only call it once.
        self._load_connection.restart_and_call(
            self.__on_load_first_use
        )
        # Now we're in a fully synced state, meaning
        # we don't need to do anything fancy with the load connection
        # anymore. Let it know that.
        self._load_connection.active = True

        changed_oids = self.__queued_changes
        self.__queued_changes = OID_SET_TYPE()

        if changed_oids is None:
            oids = None
        else:
            # The Connection doesn't care about this, it just passes it
            # to the PickleCache. The Python implementation of the PickleCache
            # takes any iterable of oid bytes, but the C implementation deals only
            # with an actual dict(), or something that is a sequence and can be
            # iterated using integer indices. If you give it a dict, all it cares
            # about are the keys.
            oids = {int64_to_8bytes(oid_int): 1 for oid_int in changed_oids}
        return oids

    def __stale(self, stale_error):
        # Reset the Connection cache. This makes it more likely we'll be called to
        # load an object and the application can notice the issue.
        self.__queued_changes = None
        # Allow GC to do its thing with the locals, don't hold them indefinitely
        clear_frames(sys.exc_info()[2])

        replacements = {}
        my_ns = vars(self)
        for k, v in my_ns.items():
            if callable(getattr(v, 'stale', None)):
                new_v = v.stale(stale_error)
                replacements[k] = new_v

        my_ns.update(replacements)

        # The next time we rollback or reopen, we're no longer stale
        # Note that this creates a reference cycle, but it should only be
        # temporary.
        self._load_connection.on_rolledback = self.__no_longer_stale
        self._load_connection.on_opened = self.__no_longer_stale

    def __no_longer_stale(self, _conn, _cursor):
        # This is called at the end of every transaction in
        # afterCompletion(), and at the beginning of every transaction
        # in sync() (called from Connection.newTransaction(), which is
        # called automatically from Connection.afterCompletion when
        # transactions are not explicit) Most of the time we won't be
        # stale, which is why we only install these hooks when we are,
        # and are careful to clean them up when we're not.
        #
        # Both _conn and _cursor will be None if the rollback detected
        # a disconnect from the database.

        my_ns = vars(self)
        replacements = {
            k: v.no_longer_stale()
            for k, v in my_ns.items()
            if callable(getattr(v, 'no_longer_stale', None))
        }

        my_ns.update(replacements)

        del self._load_connection.on_rolledback
        del self._load_connection.on_opened

    def __on_load_first_use(self, conn, cursor):
        """
        Poll for invalidations, update our cache.

        Move the component objects of this object into or out of the
        stale state as appropriate.

        Primarily in tests which directly use storage APIs, this may
        be called automatically, between transactions, when a
        Connection is not in use to receive invalidations. For that
        reason, we queue changes seen here as we keep moving forward
        and can apply them all when a connection asks us.
        """

        # Ignore changes made by the last transaction committed
        # by this connection, we don't want to ghost objects that we're sure
        # are up-to-date unless someone else has changed them.
        # Note that transactions can happen between us committing and polling.
        if self._last_tid_i_committed_bytes is not None:
            ignore_tid = bytes8_to_int64(self._last_tid_i_committed_bytes)
        else:
            ignore_tid = None

        # get a list of changed OIDs and the most recent tid
        try:
            changes = self._cache.poll(
                conn, cursor,
                ignore_tid
            )
        except (ReadConflictError, CacheConsistencyError) as e:
            # The database connection or cache is stale, but postpone
            # this error until the application tries to read or write
            # something and is better equipped to deal with it. (This
            # is called from transaction.begin() ->
            # Connection.newTransaction ->
            # RelStorage.poll_invalidations, and no one is prepared
            # for transaction.begin() to fail, or to handle it
            # gracefully if it does.)

            # If we were configured to revert-when-stale, we wouldn't
            # get a ReadConflictError, we'd have a return of None and
            # our cache would restart polling. Eventually the RCE
            # should stop as the replica catches up. But if we clear
            # the caches on the RCE, they'll lose their polling state
            # and essentially revert-when-stale even though we weren't
            # configured to do so.

            # At a minimum, though, we must reset our Connection's
            # object cache. If we fail do do this, then we see issue
            # in the zodbconvert tests, when the destination storage
            # configuration loaded to do the conversion and zapping on
            # didn't have matching persistent cache information as the
            # storage configuration that had initially populated the
            # destination (in order to verify zapping). As our cache
            # implementation got better and better, we suddenly found
            # that without zapping the cache files, we would report
            # this RCE, based on info from the persistent cache from
            # before the conversion, with no way to fix it (short of
            # zapping those cache files). This can also come up with
            # packing in history-free databases
            # (checkPackAllRevisions). We have a temporary fix for
            # that in self.pack() until we write more specific tests.
            #
            # TODO: We need a way to distinguish RCE from packing
            # vs RCE from a replica switch.

            # A CCE, on the other hand, means the cache has already
            # reset its own internal state, we outght to do the same for
            # the Connection. See above for why this shouldn't
            # propagate.
            log.error("Reading from stale replica or leading transactions packed away; %s", e)
            self.__stale(e)
            changes = None


        if changes is None:
            # Stop tracking invalidations and reset the ZODB object cache.
            # This is reset by poll_invalidations.
            self.__queued_changes = None
        elif self.__queued_changes is not None:
            self.__queued_changes.update(changes)
            if len(self.__queued_changes) > self._options.cache_delta_size_limit:
                # Hmm, ok, the Connection isn't polling us in a timely fashion.
                # Maybe we're the root storage? Maybe our APIs are being used
                # independently? At any rate, we're going to stop tracking now;
                # if a Connection eventually gets around to polling us, they'll
                # need to clear their whole cache
                self.__queued_changes = None

    def temporaryDirectory(self):
        """Return a directory that should be used for uncommitted blob data.

        If Blobs use this, then commits can be performed with a simple rename.
        """
        return self.blobhelper.temporaryDirectory()

    def copyTransactionsFrom(self, other):
        Copy(self.blobhelper, self, self).copyTransactionsFrom(other)
        self._adapter.stats.large_database_change()

    def pack(self, t, referencesf, prepack_only=False, skip_prepack=False):
        pack = Pack(self._options, self._adapter, self.blobhelper, self._cache)
        pack.pack(t, referencesf, prepack_only, skip_prepack)
        if not self.keep_history:
            # In a history free database, it's *possible*
            # that the database's last transaction ID could now have actually
            # gone backwards! I strongly suspect this only happens in fabricated
            # test scenarios (e.g., the last transaction in the database is an object that
            # was directly stored without any connection to the object graph starting from the
            # root and hence was packed away). Not only do we need to restart our
            # connection, we also need to restart our polling. Moreover, to prevent
            # loadSerial() from finding cached data, we also need to flush our caches.
            self._cache.clear(load_persistent=False)

        self.sync()

        self._pack_finished()

    def _pack_finished(self):
        "Hook for testing."


def _zlibstorage_new_instance(self):
    new_self = type(self).__new__(type(self))
    # Preserve _transform, etc
    new_self.__dict__ = self.__dict__.copy()
    new_self.base = self.base.new_instance()
    # Because these are bound methods, we must re-copy
    # them or ivars might be wrong, like _transaction
    for name in self.copied_methods:
        v = getattr(new_self.base, name, None)
        if v is not None:
            setattr(new_self, name, v)
    return new_self

def _zlibstorage_pack(self, pack_time, referencesf, *args, **kwargs):
    untransform = self._untransform
    def refs(state, oids=None):
        return referencesf(untransform(state), oids)
    return self.base.pack(pack_time, refs, *args, **kwargs)

def _zlibstorage_Iterator_len(self):
    return len(self._base_it)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path

from zope.interface import implementer

from ..interfaces import ISchemaInstaller
from ..schema import AbstractSchemaInstaller


@implementer(ISchemaInstaller)
class Sqlite3SchemaInstaller(AbstractSchemaInstaller):

    COLTYPE_OID_TID = 'INTEGER'
    COLTYPE_BINARY_STRING = 'BLOB'
    COLTYPE_STATE = 'BLOB'
    COLTYPE_BLOB_CHUNK_NUM = 'INTEGER'
    COLTYPE_BLOB_CHUNK = 'BLOB'
    COLTYPE_MD5 = 'BLOB'

    all_tables = tuple(
        t for t in AbstractSchemaInstaller.all_tables
        if t != 'new_oid'
    )

    def __init__(self, driver, oid_allocator, **kwargs):
        self.driver = driver
        self.oid_allocator = oid_allocator
        super(Sqlite3SchemaInstaller, self).__init__(**kwargs)

    def get_database_name(self, cursor):
        return os.path.splitext(
            os.path.basename(self.connmanager.path))[0]

    def list_tables(self, cursor):
        cursor.execute(
            'SELECT name FROM sqlite_master '
            'WHERE type = "table"'
        )
        return [x[0] for x in cursor.fetchall()]

    def list_views(self, cursor):
        cursor.execute(
            'SELECT name FROM sqlite_master '
            'WHERE type = "view"'
        )
        return [x[0] for x in cursor.fetchall()]

    def list_procedures(self, cursor):
        return ()

    list_sequences = list_procedures

    def _create_pack_lock(self, cursor):
        """Does nothing."""

    def _reset_oid(self, cursor):
        self.oid_allocator.reset_oid()

    def drop_all(self):
        super(Sqlite3SchemaInstaller, self).drop_all()
        self.oid_allocator.reset_oid()



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import unittest

from zope.interface import implementer

from ..interfaces import DriverNotAvailableError
from ..interfaces import NoDriversAvailableError
from ..interfaces import UnknownDriverError
from ..interfaces import IDBDriverOptions

from .. import drivers as abstract_drivers

@implementer(IDBDriverOptions)
class MockDrivers(object):
    database_type = 'mock'

    def __init__(self):
        self.driver_order = []

    def known_driver_factories(self):
        return self.driver_order

    def select_driver(self, driver_name=None):
        return abstract_drivers._select_driver_by_name(driver_name, self)

MockFactory = abstract_drivers._ClassDriverFactory

class TestAbstractDrivers(unittest.TestCase):

    def test_mock(self):
        from hamcrest import assert_that
        from nti.testing.matchers import validly_provides
        drivers = MockDrivers()
        assert_that(drivers, validly_provides(IDBDriverOptions))

    def test_select_auto_no_drivers(self):
        drivers = MockDrivers()
        with self.assertRaises(NoDriversAvailableError):
            drivers.select_driver()

    def test_select_auto(self):
        drivers = MockDrivers()
        drivers.driver_order.append(MockFactory(lambda: 42))
        d = drivers.select_driver()
        self.assertEqual(d, 42)

    def test_select_auto_not_available(self):
        class BadDriver(object):
            def __init__(self):
                raise DriverNotAvailableError('Bad')

        drivers = MockDrivers()
        drivers.driver_order.append(MockFactory(BadDriver))

        with self.assertRaises(NoDriversAvailableError) as exc:
            drivers.select_driver()

        self.assertIn(
            "Driver 'auto' is not available "
            "(reason={'BadDriver': \"DriverNotAvailableError: Driver 'Bad' is not available.\"}). "
            "Options: 'BadDriver' (Module: '<unknown>'; Available: False).",
            str(exc.exception))

    def test_select_unknown_name(self):
        drivers = MockDrivers()
        with self.assertRaises(UnknownDriverError):
            abstract_drivers._select_driver_by_name('DNE', drivers)

    def test_case_insensitive_names(self):
        drivers = MockDrivers()

        class Driver(object):
            __name__ = 'MixedCase'

        drivers.driver_order.append(MockFactory(Driver))

        self.assertIsInstance(drivers.select_driver('mixedcase'), Driver)
        self.assertIsInstance(drivers.select_driver('MIXEDCASE'), Driver)
        self.assertIsInstance(drivers.select_driver('MixedCase'), Driver)


class IDBDriverSupportsCriticalTestMixin(object):

    def _makeOne(self):
        raise NotImplementedError

    def test_enter_critical_phase(self):
        driver = self._makeOne()
        conn = driver.connect()

        self.assertFalse(conn.is_in_critical_phase())
        driver.enter_critical_phase_until_transaction_end(conn, None)
        self.assertTrue(driver.is_in_critical_phase(conn, None))
        # idempotent
        conn.enter_critical_phase_until_transaction_end()
        self.assertTrue(conn.is_in_critical_phase())

    def test_enter_critical_phase_exit_on_commit(self):
        driver = self._makeOne()
        conn = driver.connect()

        # exits on commit
        driver.enter_critical_phase_until_transaction_end(conn, None)
        conn.commit()
        self.assertFalse(driver.is_in_critical_phase(conn, None))

    def test_enter_critical_phase_exit_on_rollback(self):
        driver = self._makeOne()
        conn = driver.connect()

        # exits on commit
        driver.enter_critical_phase_until_transaction_end(conn, None)
        conn.rollback()
        self.assertFalse(driver.is_in_critical_phase(conn, None))



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint:disable=too-many-lines

import logging
import os
import threading

from persistent.timestamp import TimeStamp
from ZODB.POSException import ReadConflictError
from ZODB.utils import p64
from zope import interface

from relstorage._compat import IN_TESTRUNNER
from relstorage._compat import OID_SET_TYPE as OIDSet
from relstorage._compat import OID_TID_MAP_TYPE as OidTMap
from relstorage._util import bytes8_to_int64
from relstorage._mvcc import DetachableMVCCDatabaseViewer

from relstorage.cache import persistence
from relstorage.cache.interfaces import IStorageCache
from relstorage.cache.interfaces import IPersistentCache
from relstorage.cache.interfaces import CacheConsistencyError
from relstorage.cache.local_client import LocalClient
from relstorage.cache.memcache_client import MemcacheStateCache
from relstorage.cache.trace import ZEOTracer
from relstorage.cache._statecache_wrappers import MultiStateCache
from relstorage.cache._statecache_wrappers import TracingStateCache
from relstorage.cache.mvcc import MVCCDatabaseCoordinator

logger = log = logging.getLogger(__name__)

class _UsedAfterRelease(object):
    size = limit = 0
    def __len__(self):
        return 0
    def __call__(self):
        raise NotImplementedError
    close = reset_stats = release = unregister = lambda self, *args: None
    stats = lambda s: {}
    new_instance = lambda s: s
_UsedAfterRelease = _UsedAfterRelease()


@interface.implementer(IStorageCache, IPersistentCache)
class StorageCache(DetachableMVCCDatabaseViewer):
    """RelStorage integration with memcached or similar.

    Holds a list of memcache clients in order from most local to
    most global.  The first is a LocalClient, which stores the cache
    in the Python process, but shares the cache between threads.
    """
    # pylint:disable=too-many-instance-attributes,too-many-public-methods

    __slots__ = (
        'adapter',
        'options',
        'keep_history',
        'prefix',
        'polling_state',
        'local_client',
        'cache',
        'object_index',
    )


    if IN_TESTRUNNER:
        class MVCCInternalConsistencyError(Exception):
            "This can never be raised or caught."
    else:
        MVCCInternalConsistencyError = AssertionError

    def __init__(self, adapter, options, prefix, _parent=None):
        super(StorageCache, self).__init__()
        self.adapter = adapter
        self.options = options
        self.keep_history = options.keep_history
        self.prefix = prefix or ''

        if _parent is None:
            # I must be the master!

            # This is shared between all instances of a cache in a tree,
            # including the master, so that they can share information about
            # polling.
            self.polling_state = MVCCDatabaseCoordinator(self.options)
            self.local_client = LocalClient(options, self.prefix)


            shared_cache = MemcacheStateCache.from_options(options, self.prefix)
            if shared_cache is not None:
                self.cache = MultiStateCache(self.local_client, shared_cache)
            else:
                self.cache = self.local_client

            tracefile = persistence.trace_file(options, self.prefix)
            if tracefile:
                tracer = ZEOTracer(tracefile)
                tracer.trace(0x00)
                self.cache = TracingStateCache(self.cache, tracer)
        else:
            self.polling_state = _parent.polling_state # type: MVCCDatabaseCoordinator
            self.local_client = _parent.local_client.new_instance()
            self.cache = _parent.cache.new_instance()

        # Once we have registered with the MVCCDatabaseCoordinator,
        # we cannot make any changes to our own mvcc state without
        # letting it know about them. In particular, that means we must
        # not just assign to this object (except under careful circumstances
        # where we're sure to be single threaded.)
        # This object can be None
        self.object_index = None

        # It is also important not to register with the coordinator until
        # we are fully initialized; we could be constructing a new_instance
        # in a separate thread while polling is going on in other threads.
        # We can get strange AttributeError if a partially constructed instance
        # is exposed.
        self.polling_state.register(self)

        if _parent is None:
            self.restore()


    @property
    def current_tid(self):
        # testing
        return self.highest_visible_tid

    # XXX: Note that our __bool__ and __len__ are NOT consistent
    def __bool__(self):
        return True
    __nonzero__ = __bool__

    def __len__(self):
        return len(self.local_client)

    @property
    def size(self):
        return self.local_client.size

    @property
    def limit(self):
        return self.local_client.limit

    def stats(self):
        """
        Return stats. This is a debugging aid only. The format is undefined and intended
        for human inspection only.
        """
        stats = self.local_client.stats()
        stats['local_index_stats'] = self.object_index.stats() if self.object_index else None
        stats['global_index_stats'] = self.polling_state.stats()
        return stats

    def __repr__(self):
        return "<%s at 0x%x hvt=%s bytes=%d len=%d state=%r>" % (
            self.__class__.__name__,
            id(self),
            self.highest_visible_tid,
            self.size,
            len(self),
            self.polling_state,
        )

    def reset_stats(self):
        self.local_client.reset_stats()

    def new_instance(self, before=None, adapter=None):
        """
        Return a copy of this instance sharing the same local client
        and having the most current view of the database as collected
        by any instance.

        If *before* is given, the new cache will use a distinct
        :class:`MVCCDatabaseCoordinator`  so that
        its usage pattern does not interfere.
        """
        klass = type(self) if before is None else _BeforeStorageCache
        cache = klass(adapter or self.adapter,
                      self.options, self.prefix,
                      _parent=self)
        return cache

    def release(self):
        """
        Release resources held by this instance.

        This does not corrupt shared state, and must be called
        on each instance that's not the root.

        This is usually memcache connections if they're in use.
        """
        self.cache.release()
        # Release our clients. If we had a non-shared local cache,
        # this will also allow it to release any memory it's holding.
        self.local_client = self.cache = _UsedAfterRelease
        self.polling_state.unregister(self)
        self.polling_state = _UsedAfterRelease
        self.object_index = None
        self.highest_visible_tid = None

    def close(self, **save_args):
        """
        Release resources held by this instance, and
        save any persistent data necessary.

        This is only called on the root. If there are still instances
        that haven't been released, they'll be broken.
        """
        # grab things that will be reset in release()
        cache = self.cache
        polling_state = self.polling_state

        # Go ahead and release our polling_state now, in case
        # it helps to vacuum for save.
        self.polling_state.unregister(self)
        self.save(**save_args)
        self.release()
        cache.close()
        polling_state.close()

    def save(self, **save_args):
        """
        Store any persistent client data.
        """
        if self.options.cache_local_dir and len(self) > 0: # pylint:disable=len-as-condition
            # (our __bool__ is not consistent with our len)
            stats = self.local_client.stats()
            if stats['hits'] or stats['sets']:
                # Only write this out if (1) it proved useful OR (2)
                # we've made modifications. Otherwise, we're writing a consolidated
                # file for no good reason.
                # TODO: Consider the correctness here, now that we have a
                # more accurate cache. Should that maybe be AND?
                return self.polling_state.save(self, save_args)
            logger.debug("Cannot justify writing cache file, no hits or misses")

    def restore(self):
        # We must only restore into an empty cache.
        state = self.polling_state
        assert not self.local_client
        state.restore(self.adapter, self.local_client)

    def _reset(self, message=None):
        """
        Reset the transaction state of only this instance.

        If this is being done in a transactional way, it must be followed
        by raising an exception. If the *message* parameter is provided,
        then a ``CacheConsistencyError`` will be raised when this
        method returns.
        """
        # As if we've never polled
        self.polling_state.reset_viewer(self)
        self.polling_state.flush_all()
        if message:
            raise CacheConsistencyError(message)

    def clear(self, load_persistent=True):
        """
        Remove all data from the cache, both locally (and shared among
        other instances), and globally.

        Called by speed tests.

        Starting from the introduction of persistent cache files, this
        also results in the local client being repopulated with the
        current set of persistent data. The *load_persistent* keyword
        can be used to control this.

        .. versionchanged:: 2.0b6 Added the ``load_persistent``
           keyword. This argument is provisional.
        """
        self._reset()
        self.polling_state.flush_all()
        self.cache.flush_all()

        if load_persistent:
            self.restore()

    def zap_all(self):
        """
        Remove all data from the cache, both locally (and shared among
        other instances, and globally); in addition, remove any
        persistent cache files on disk.
        """
        self.local_client.zap_all()
        self.clear(load_persistent=False)

    def _check_tid_after_load(self, oid_int, actual_tid_int,
                              expect_tid_int=None,
                              cursor=None):
        """Verify the tid of an object loaded from the database is sane."""
        if actual_tid_int is not None and actual_tid_int > self.highest_visible_tid:
            # Strangely, the database just gave us data from a future
            # transaction. We can't give the data to ZODB because that
            # would be a consistency violation. However, the cause is
            # hard to track down, so issue a ReadConflictError and
            # hope that the application retries successfully.
            msg = ("Got data for OID 0x%(oid_int)x from "
                   "future transaction %(actual_tid_int)d (%(got_ts)s).  "
                   "Current transaction is %(hvt)s (%(current_ts)s)."
                   % {
                       'oid_int': oid_int,
                       'actual_tid_int': actual_tid_int,
                       'hvt': self.highest_visible_tid,
                       'got_ts': str(TimeStamp(p64(actual_tid_int))),
                       'current_ts': str(TimeStamp(p64(self.highest_visible_tid))),
                   })
            raise ReadConflictError(msg)

        if expect_tid_int is not None and actual_tid_int != expect_tid_int:
            # Uh-oh, the cache is inconsistent with the database.
            # We didn't get a TID from the future, but it's not what we
            # had in our delta_after0 map, which means...we missed a change
            # somewhere.
            #
            # Possible causes:
            #
            # - The database MUST provide a snapshot view for each
            #   session; this error can occur if that requirement is
            #   violated. For example, MySQL's MyISAM engine is not
            #   sufficient for the object_state table because MyISAM
            #   can not provide a snapshot view. (InnoDB is
            #   sufficient.)
            #
            # - (Similar to the last one.) Using too low of a
            #   isolation level for the database connection and
            #   viewing unrelated data.
            #
            # - Something could be writing to the database out
            #   of order, such as a version of RelStorage that
            #   acquires a different commit lock.
            #
            # - A software bug. In the past, there was a subtle bug
            #   in after_poll() that caused it to ignore the
            #   transaction order, leading it to sometimes put the
            #   wrong tid in delta_after*.
            #
            # - Restarting a load connection at a future point we hadn't
            #   actually polled to, such that our current_tid is out of sync
            #   with the connection's *actual* viewable tid?
            from pprint import pformat
            from relstorage._util import int64_to_8bytes
            msg = ("Detected an inconsistency "
                   "between the RelStorage cache and the database "
                   "while loading an object using the MVCC index.  "
                   "Please verify the database is configured for "
                   "ACID compliance and that all clients are using "
                   "the same commit lock. Info:\n%s"
                   % pformat({
                       'oid_int': oid_int,
                       'expect_tid_int': expect_tid_int,
                       'actual_tid_int': actual_tid_int,
                       # Typically if this happens we get something newer than we expect.
                       'actual_expect_delta': actual_tid_int - expect_tid_int,
                       'expect_tid': str(TimeStamp(int64_to_8bytes(expect_tid_int))),
                       'actual_tid': str(TimeStamp(int64_to_8bytes(actual_tid_int))),
                       'current_tid': self.highest_visible_tid,
                       'pid': os.getpid(),
                       'thread_ident': threading.current_thread(),
                       'cursor': cursor,
                   }))
            # We reset ourself as if we hadn't polled, and hope the transient
            # error gets retried in a working, consistent view.
            self._reset(msg)

    def loadSerial(self, oid_int, tid_int):
        """
        Return the locally cached state for the object *oid_int* as-of
        exactly *tid_int*.

        If that state is not available in the local cache, return
        nothing.

        This is independent of the current transaction and polling state, and
        may return data from the future.

        If the storage hasn't polled invalidations, or if there are other viewers
        open at transactions in the past, it may also return data from the past
        that has been overwritten (in history-free storages).
        """
        # We use only the local client because, for history-free storages,
        # it's the only one we can be reasonably sure has been
        # invalidated by a local pack. Also, our point here is to avoid
        # network traffic, so it's no good going to memcache for what may be
        # a stale answer.

        cache = self.local_client
        # Don't take this as an MRU hit; if we succeed, we'll
        # put new cached data in for this OID and do that anyway.
        cache_data = cache.get((oid_int, tid_int), False)
        if cache_data and cache_data[1] == tid_int:
            return cache_data[0]

    def load(self, cursor, oid_int):
        """
        Load the given object from cache if possible.

        Fall back to loading from the database.

        Returns (state_bytes, tid_int).
        """
        # pylint:disable=too-many-statements,too-many-branches,too-many-locals
        if not self.object_index:
            # No poll has occurred yet. For safety, don't use the cache.
            # Note that without going through the cache, we can't
            # go through tracing either.
            return self.adapter.mover.load_current(cursor, oid_int)

        # Get the object from the transaction specified
        # by the following values, in order:
        #
        #   1. self.object_index[oid_int]
        #
        # An entry in object_index means we've polled for and know the exact
        # TID for this object, either because we polled, or because someone
        # loaded it and put it in the index. If we know a TID, we must *never*
        # use the wildcard frozen value (it's possible to have an older frozen tid that's
        # valid for older transactions, but out of date for this one.) That's handled
        # internally in the clients.


        cache = self.cache
        index = self.object_index
        indexed_tid_int = index[oid_int] # Could be None pylint:disable=unsubscriptable-object

        key = (oid_int, indexed_tid_int)
        cache_data = cache[key]
        if cache_data and indexed_tid_int is None and cache_data[1] > self.highest_visible_tid:
            # Cache hit on a wildcard, but we need to verify the wildcard
            # and it didn't pass. This situation should be impossible.
            cache_data = None

        if cache_data:
            # Cache hit, non-wildcard or wildcard matched.
            return cache_data

        # Cache miss.
        state, actual_tid_int = self.adapter.mover.load_current(
            cursor, oid_int)
        if actual_tid_int:
            # If either is None, the object was deleted.
            self._check_tid_after_load(oid_int, actual_tid_int, indexed_tid_int, cursor)

            # We may or may not have had an index entry, but make sure we do now.
            # Eventually this will age to be frozen again if needed.
            index[oid_int] = actual_tid_int # pylint:disable=unsupported-assignment-operation
            cache[(oid_int, actual_tid_int)] = (state, actual_tid_int)
            return state, actual_tid_int

        # This is in the bytecode as a LOAD_CONST
        return None, None

    def prefetch(self, cursor, oid_ints):
        # Just like load(), but we only fetch the OIDs
        # we can't find in the cache.
        if not self.object_index:
            # No point even trying, we would just throw the results away
            return

        cache = self.cache
        if cache is self.local_client and not cache.limit:
            # No point.
            return

        index = self.object_index
        # We don't actually need the cache data, so avoid asking
        # for it. That would trigger stats updates (hits/misses)
        # and move it to the front of the LRU list. But this is just
        # in advance, we don't know if it will actually be used.
        # `in` has a race condition (it could be evicted soon), but
        # if it is, there was probably something else more important
        # going on.
        to_fetch = {oid_int
                    for oid_int in oid_ints
                    if (oid_int, index[oid_int]) not in cache} # pylint:disable=unsubscriptable-object

        if not to_fetch:
            return

        for oid, state, tid_int in self.adapter.mover.load_currents(cursor, to_fetch):
            key = (oid, tid_int)
            self._check_tid_after_load(oid, tid_int, cursor=cursor)
            cache[key] = (state, tid_int)
            index[oid] = tid_int # pylint:disable=unsupported-assignment-operation

    def prefetch_for_conflicts(self, cursor, oid_tid_pairs):
        results = {}
        to_fetch = OidTMap()
        cache_get = self.cache.get

        # if we've never polled, we can't actually use our cache, will
        # just have to make a bulk query.
        if not self.object_index:
            to_fetch = OidTMap(oid_tid_pairs)
        else:
            for key in oid_tid_pairs:
                # Don't update stats/MRU, just as with normal prefetch().
                # It's also important here to avoid taking the lock.
                # We don't store this prefetched data back into the cache because
                # we're just about to overwrite it; we'd have to have multiple writers
                # all with the same initial starting TID lined up to write to the object
                # for that to have any benefit.
                cache_data = cache_get(key, peek=True)
                if not cache_data:
                    to_fetch[key[0]] = key[1]
                else:
                    results[key[0]] = cache_data
                    assert cache_data[1] == key[1]

        if to_fetch:
            check = self._check_tid_after_load if self.object_index else lambda *_, **kw__: None
            for oid, state, tid_int in self.adapter.mover.load_currents(cursor, to_fetch):
                check(oid, tid_int, to_fetch[oid], cursor=cursor)
                results[oid] = (state, tid_int)

        return results


    def remove_cached_data(self, oid_int, tid_int):
        """
        See notes in `invalidate_all`.
        """
        del self.cache[(oid_int, tid_int)]

    def remove_all_cached_data_for_oids(self, oids):
        """
        Invalidate all cached data for the given OIDs.

        This isn't transactional or locked so it may still result in
        this or others seeing invalid (ha!) states.

        This is a specialized API. It allows violation of our internal
        consistency constraints. It should only be used when the
        database is being manipulated at a low level, such as during
        pack or undo.
        """
        # Erase our knowledge of where to look
        # self._invalidate_all(oids)
        # Remove the data too.
        self.cache.invalidate_all(oids)

    def after_tpc_finish(self, tid, temp_storage):
        """
        Flush queued changes.

        This is called after the database commit lock is released,
        but before control is returned to the Connection.

        Now that this tid is known, send all queued objects to the
        cache. The cache will have ``(oid, tid)`` entry for each object
        we have been holding on to (well, in a big transaction, some of them
        might actually not get stored in the cache. But we try!)
        """
        tid_int = bytes8_to_int64(tid)
        self.cache.set_all_for_tid(tid_int, temp_storage)

    def poll(self, conn, cursor, ignore_tid):
        try:
            changes = self.polling_state.poll(self, conn, cursor)
        except self.MVCCInternalConsistencyError: # pragma: no cover
            logger.critical(
                "Internal consistency violation in the MVCC coordinator. "
                "Please report a bug to the RelStorage maintainers. "
                "Flushing caches for safety. ",
                exc_info=True
            )
            self._reset("Unknown internal violation")

        if changes is not None:
            return OIDSet(oid for oid, tid in changes if tid != ignore_tid)


class _BeforeStorageCache(StorageCache):

    __slots__ = ()

    def poll(self, conn, cursor, _):
        # Grab whatever index we have the first time we poll,
        # and then immediately drop away from the polling state.
        # We don't want our increasingly-stale view of the database
        # (which isn't even really quite accurate in terms of what we
        # need: TODO: Should we do that?) to hold up vacuums (even
        # though it will hold up vacuum in the real RDBMS).
        self.polling_state.poll(self, conn, cursor)
        # We'll never try to poll again, so this is ok.
        self.polling_state.unregister(self)
        return ()



##############################################################################
#
# Copyright (c) 2009,2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os
import re
import time

from binascii import hexlify

from BTrees import OOBTree # pylint:disable=no-name-in-module
import zc.lockfile

import ZODB.blob
from zope.interface import implementer

from relstorage._util import byte_display
from relstorage._util import spawn as native_thread_spawn
from relstorage._util import thread_spawn
from relstorage._util import timer
from relstorage._util import bytes8_to_int64
from relstorage.interfaces import POSKeyError

from .interfaces import ICachedBlobHelper
from .abstract import AbstractBlobHelper
from .util import lock_blob


logger = __import__('logging').getLogger(__name__)


class _AbstractCacheSizeMonitor(object):
    __slots__ = ()

    def __init__(self, options):
        """Does nothing."""

    def close(self):
        """
        Finish any processing and free any resources used by this monitor.
        """
        raise NotImplementedError

    def loaded(self, byte_count):
        """
        Let the monitor know that *byte_count* bytes of data have been
        added to the cache.
        """
        raise NotImplementedError

class _UnlimitedCacheSizeMonitor(_AbstractCacheSizeMonitor):
    """
    Use this when no limit has been configured.
    """
    __slots__ = ()

    def close(self):
        """Does nothing."""

    def loaded(self, byte_count):
        """Does nothing."""


class _LimitedCacheSizeMonitor(_AbstractCacheSizeMonitor):
    """
    Control the size of the blob cache. This object is shared
    between BlobHelpers, so it needs to be thread safe.
    """

    __slots__ = (
        'blob_dir',
        'blob_cache_max_size',
        'blob_cache_target_cleanup_size',
        'bytes_loaded_since_last_check',
        'bytes_loaded_check_threshold',
        '_lock',
        '_checker_thread',
        '_reduced_event',
        '_exceeded_counter',
    )

    def __init__(self, options):
        import threading

        assert options.blob_cache_size_check < 100

        super(_LimitedCacheSizeMonitor, self).__init__(options)

        self.blob_dir = options.blob_dir
        self.blob_cache_max_size = options.blob_cache_size
        self.bytes_loaded_check_threshold = (
            self.blob_cache_max_size * options.blob_cache_size_check / 100.0
        )

        self.blob_cache_target_cleanup_size = max(
            self.blob_cache_max_size - self.bytes_loaded_check_threshold,
            0
        )
        self.bytes_loaded_since_last_check = 0
        # thread-safe according to whatever threading system is in use.
        # We *want* mankey patches here.
        self._lock = threading.Lock()
        self._reduced_event = threading.Event()
        self._checker_thread = None
        self._exceeded_counter = 0
        self._check()

    def close(self):
        try:
            if self._checker_thread is not None:
                self.wait_for_checker()
        finally:
            self._checker_thread = None

    def loaded(self, byte_count):
        with self._lock:
            self.bytes_loaded_since_last_check += byte_count
            if self.bytes_loaded_since_last_check >= self.bytes_loaded_check_threshold:
                logger.debug(
                    "Loaded %s bytes (>= %s) into %s, may need to check.",
                    byte_display(self.bytes_loaded_since_last_check),
                    byte_display(self.bytes_loaded_check_threshold),
                    self.blob_dir
                )
                self._check()

    def wait_for_checker(self):
        """For tests only."""
        # Efficiently wait only if a checker thread is running.
        # It's native, a threading.Event() might not be and we'd need an
        # asyncwatcher to communicate among threads. The tests that use us patch
        # to be sure we use matching events.
        with self._lock:
            if self._checker_thread is None:
                return
        self._reduced_event.wait()

    def _check(self):
        """
        Run blob cache cleanup in another thread if needed.

        Must be called with our lock held (or a guarantee that we're
        single threaded.)
        """
        on_init = self.bytes_loaded_since_last_check == 0
        self.bytes_loaded_since_last_check = 0
        self._reduced_event.clear()
        self._exceeded_counter += 1
        checker_thread = self._checker_thread
        if checker_thread is not None and not checker_thread.ready():
            # One running still.
            logger.debug("Checker %s still running, not spawning for %s",
                         self._checker_thread,
                         self.blob_dir)
            return

        # Only spawn a new one if there's not one running.
        # This gets set back to None as part of the cleanup callback.
        logger.info(
            "Spawning cache checker for %s (%s)",
            self.blob_dir,
            "creating storage" if on_init else "exceeded threshold"
        )
        self._exceeded_counter = 0
        self._checker_thread = self._spawn()

    def _spawn(self):
        checker = _BlobCacheSizeChecker(
            self.blob_dir, self.blob_cache_target_cleanup_size, self._when_done
        )
        return native_thread_spawn(checker)

    def _when_done(self, checker, holding_clean_lock):
        """
        Callback to be run from the cleanup thread.

        Cleans up the state of *self*.
        """
        with self._lock:
            # checker is the BlobCacheSizeChecker, but self._checker
            # is the spawned thread.
            # This is the last thing the BlobCacheSizeChecker does, so by
            # definition it cannot be ready yet.
            self._checker_thread = None
            if not holding_clean_lock:
                self._reduced_event.set()
                return

            # In principle, if the checker finished sizing the directory and got
            # a cache size under its target and wanted to return to us,
            # but then some other threads immediately loaded a bunch of blobs,
            # we could go over that size. We prevent this by checking
            # the size again here, while we're holding our lock, and if we're
            # too big, we'll go again. This happens during the test cases.
            # checker can be None if we did this externally.
            dir_size = -1
            duration = 0.0
            if checker is not None:
                dir_size = checker.blob_dir_size
                duration = checker.duration

            logger.info(
                "Finished checking %s (in %.3fs) with size of %s (max: %s; target %s)",
                self.blob_dir, duration,
                byte_display(dir_size),
                byte_display(self.blob_cache_max_size),
                byte_display(self.blob_cache_target_cleanup_size)
            )
            if self._exceeded_counter or dir_size > self.blob_cache_target_cleanup_size:
                logger.debug(
                    "Requesting new check for %s with size of %s (max: %s; target %s)",
                    self.blob_dir,
                    dir_size,
                    self.blob_cache_max_size,
                    byte_display(self.blob_cache_target_cleanup_size)
                )
                self._check()
            else:
                self._reduced_event.set()


class _ExternalLimitedCacheSizeMonitor(_LimitedCacheSizeMonitor):
    """
    Uses a subprocess instead of a thread.
    """
    __slots__ = ()

    @staticmethod
    def __run_checker(blob_dir, blob_cache_target_cleanup_size, when_done):
        import subprocess
        import sys
        try:
            popen = subprocess.Popen([
                sys.executable,
                "-m",
                __name__,
                blob_dir,
                str(int(blob_cache_target_cleanup_size))
            ])

            popen.wait()
        finally:
            # Pretend that we held the check lock, even if we're not
            # sure that we did. If we need to go again because we've already
            # stored that much more data, that lets us schedule the process.
            # TODO: Make the external process write something we can read,
            # or use the exit code to determine if it had the lock.
            when_done(None, True)
            when_done = None

    def _spawn(self):
        # We need to call our _when_done handler so that tests can
        # determine when to look at the directory. This does not need
        # to be a native thread under gevent, a greenlet thread is
        # perfect (native threads in the threadpool can be limited),
        # but we do need to wait on the popen object.
        return thread_spawn(self.__run_checker,
                            args=(self.blob_dir,
                                  self.blob_cache_target_cleanup_size,
                                  self._when_done))


@implementer(ICachedBlobHelper)
class CacheBlobHelper(AbstractBlobHelper):

    NEEDS_DB_LOCK_TO_VOTE = False
    NEEDS_DB_LOCK_TO_FINISH = False

    def __init__(self, options, adapter, fshelper=None, cache_checker=None):
        assert not options.shared_blob_dir

        if fshelper is None:
            # The blob directory is a cache of the blobs
            if _BlobCacheLayout.LAYOUT_NAME not in ZODB.blob.LAYOUTS:
                ZODB.blob.LAYOUTS[_BlobCacheLayout.LAYOUT_NAME] = _BlobCacheLayout()
            fshelper = ZODB.blob.FilesystemHelper(
                options.blob_dir, layout_name=_BlobCacheLayout.LAYOUT_NAME)
            fshelper.create()

        super(CacheBlobHelper, self).__init__(options, adapter, fshelper)

        # All blob helpers for all instances of this storage share the
        # same cache_checker object.
        if cache_checker is None:
            factory = _UnlimitedCacheSizeMonitor
            if options.blob_cache_size:
                factory = _LimitedCacheSizeMonitor
                if options.blob_cache_size_check_external:
                    factory = _ExternalLimitedCacheSizeMonitor
            cache_checker = factory(options)

        self.cache_checker = cache_checker
        self.new_instance_kwargs['cache_checker'] = self.cache_checker

    def close(self):
        super(CacheBlobHelper, self).close()
        self.cache_checker.close()

    def _loadBlobInternal(self, cursor, oid, serial, blob_lock=None):
        blob_filename = self._cachedLoadBlobInternal(oid, serial)
        if not blob_filename:
            # OK, it's not on disk in our cache. We need to lock and
            # download. In order to lock, we need to create the directory
            # first.
            blob_filename = self._get_lockable_blob_filename(oid, serial)
            my_lock = lock_blob(blob_filename) if blob_lock is None else blob_lock
            try:
                blob_filename = self._loadBlobLocked(cursor, oid, serial, blob_filename)
            finally:
                if blob_lock is None:
                    # If we take out the lock, we close the lock.
                    # Otherwise, it's the caller's responsibility.
                    my_lock.close()
        return blob_filename

    def _loadBlobLocked(self, cursor, oid, serial, blob_filename):
        """
        Returns a filename that exists on disk, or raises a POSKeyError.
        """
        # OK, it's not here and we (or someone) needs to get it. We
        # want to avoid getting it multiple times. We want to avoid
        # getting it multiple times even accross separate client
        # processes on the same machine. We'll use file locking.
        # (accomplished by our caller.)

        # We got the lock, so it's our job to download it. First,
        # we'll double check that someone didn't download it while
        # we were getting the lock:
        if os.path.exists(blob_filename):
            return self._accessed(blob_filename)

        self.download_blob(cursor, oid, serial, blob_filename)

        if os.path.exists(blob_filename):
            return self._accessed(blob_filename)
        __traceback_info__ = cursor
        raise POSKeyError(oid, serial=serial, fn=blob_filename)

    def upload_blob(self, cursor, oid, serial, filename):
        """
        Upload a blob from a file.

        If serial is None, upload to the temporary table.
        """
        if serial is not None:
            tid_int = bytes8_to_int64(serial)
        else:
            tid_int = None
        self.adapter.mover.upload_blob(cursor, bytes8_to_int64(oid), tid_int, filename)

    def download_blob(self, cursor, oid, serial, filename):
        """Download a blob into a file"""
        tmp_fn = filename + ".tmp"
        bytecount = self.adapter.mover.download_blob(
            cursor, bytes8_to_int64(oid), bytes8_to_int64(serial), tmp_fn)
        if os.path.exists(tmp_fn):
            os.rename(tmp_fn, filename)
        self.cache_checker.loaded(bytecount)

    def storeBlob(self, cursor, store_func,
                  oid, serial, data, blobfilename, version, txn):
        assert not version
        temp_path = self._doStoreBlob(
            store_func,
            oid, serial, data, blobfilename,
            txn
        )
        self.upload_blob(cursor, oid, None, temp_path)

    def restoreBlob(self, cursor, oid, serial, blobfilename):
        # This bypasses ``txn_has_blobs``. That shouldn't cause any problems
        # because ``serial`` should not be none here, so we are avoiding the
        # temp tables. The use of ``txn_has_blobs`` in tpc_vote() is to check if we need
        # to move temporary blobs into place.
        self.upload_blob(cursor, oid, serial, blobfilename)

    def copy_undone(self, copied, tid):
        """
        Not needed in a cache.
        """

    def after_pack(self, oid_int, tid_int):
        """
        Not needed in a cache.

        Although, it might be helpful as a size control?
        """

    def _remove_old_revisions_of_stored_blobs(self, tid, total_size_stored):
        """
        Prune old revisions of blobs that are not in use.

        This is only done if we're not keeping history. This assumes
        the _BlobCacheLayout. (TODO: Generalize so we can do this for
        shared blob dirs? That's slightly more dangerous since it's
        our only copy of the data.)

        Because this could reduce the disk footprint of the cache, it
        takes the total amount of data stored, and returns a modified
        value (less if we removed old revisions).
        """
        if not total_size_stored or self.options.keep_history:
            return total_size_stored

        # If we've added/edited some blobs in this transaction,
        # and we're not keeping history, see if we can clean up
        # some older blobs for the same OIDs. This will help keep
        # our cache size contained.
        blob_suffix = ZODB.blob.BLOB_SUFFIX
        get_dir_for_oid = self.fshelper.getPathForOID # Including the base dir
        get_local_path_for_oid_tid = self.fshelper.layout.getBlobFilePath
        total_size = total_size_stored

        for stored_blob_oid in self._txn_blobs:
            # Chop off the first part of the OID; that's implicit in the full path
            # that we pass to listdir()
            stored_blob_file_path = os.path.split(
                get_local_path_for_oid_tid(stored_blob_oid, tid))[1]
            stored_oid_part, stored_tid_part, _ = stored_blob_file_path.split('.')

            dir_for_oid = get_dir_for_oid(stored_blob_oid)
            all_blob_files = [fname
                              for fname in os.listdir(dir_for_oid)
                              if fname.endswith(blob_suffix)]
            if len(all_blob_files) < 2:
                # Nothing to do if there's only one file.
                continue

            for filename in all_blob_files:
                # TODO: We could save some calls to listdir() if we collected
                # all the OIDs in self._txn_blobs that will wind up sharing
                # a single directory. That would be a lot of blobs in a transaction,
                # though.

                # The "right" way to get an oid and tid from a path is to use
                # fshelper.splitBlobFilename(), but it assumes that the first part of the
                # path component contains the entire OID, whereas here we only have
                # the remainder.
                disk_oid_part, disk_tid_part, _ = filename.split('.')
                if disk_oid_part != stored_oid_part:
                    continue

                filepath = os.path.join(self.blob_dir, dir_for_oid, filename)
                if disk_tid_part < stored_tid_part:
                    # The TID is stored in hex form. The hex form sorts identically to the
                    # byte or integer form, with increasing values meaning newer tids.
                    logger.debug(
                        "Found older cached version of the blob %s at %s; attempting removal.",
                        stored_blob_oid, filepath
                    )
                    # This takes the lock and removes it.
                    total_size -= _BlobCacheSizeChecker.remove_blob_at_path(filepath)

        # It's possible we actually removed more than we stored, if lots of them
        # were still open before.
        return total_size if total_size >= 0 else 0

    def finish(self, tid):
        try:
            total_size = self._move_blobs_into_place(tid)
            # If this slows commit down too much, we could push it to a thread
            # in a few different ways (a queue.Queue consumer, or just spawn())
            total_size = self._remove_old_revisions_of_stored_blobs(tid, total_size)
            self.cache_checker.loaded(total_size)
        except Exception: # pylint:disable=broad-except
            # We're a cache, we can ignore issues moving things into
            # place or cleaning up old revisions. The data is still
            # safe and will be downloaded if needed. Raising exceptions from
            # finish() is bad, so don't do that.
            logger.exception("Failed to properly put blob cache files into place.")
        finally:
            super(CacheBlobHelper, self).finish(tid)


class _BlobCacheLayout(object):
    """
    Uses a two-level directory layout::

        <blob-dir>/<oid1>/<oid2>.<tid>.blob

    For example::

        <blob-dir>/23/0.03d167f919308700.blob

    The ``<oid1>`` (directory name) and ``<oid2>`` (first part of the
    filename) are derived from the OID of the blob when treated as a
    64-bit integer; ``<oid1>`` will only ever contain one or more
    ASCII digits.
    """

    # This layout is a clone of the ZEO.ClientStorage.BlobCacheLayout
    # class. We haven't changed anything about how it is structured,
    # but we *might* in the future; we'd like to change the name, but
    # that would invalidate all existing caches (the layout name is
    # stored in a file on disk and checked when the FilesystemHelper is
    # created).
    #
    # TODO: In particular, even though a history-free storage only has
    # one revision of a blob in the database, we don't consider that
    # when we're caching a blob, or when we're cleaning blobs up. We
    # should be able to do better.
    LAYOUT_NAME = 'zeocache'

    size = 997

    def oid_to_path(self, oid):
        rem = bytes8_to_int64(oid) % self.size
        return str(rem)

    def getBlobFilePath(self, oid, tid):
        base, rem = divmod(bytes8_to_int64(oid), self.size)
        return os.path.join(
            str(rem),
            "%s.%s%s" % (
                base,
                hexlify(tid).decode('ascii'),
                ZODB.blob.BLOB_SUFFIX
            )
        )

class _BlobCacheSizeChecker(timer):

    __slots__ = (
        'blob_dir',
        # The last measured size of the blob directory.
        'blob_dir_size',
        'target_size',
        '_finished_callback',
        # To name a thread
        '__name__',
        'duration',
    )

    def __init__(self, blob_dir, target_size, when_done=lambda _me, _holding_lock: None):
        with open(os.path.join(blob_dir, ZODB.blob.LAYOUT_MARKER)) as layout_file:
            layout = layout_file.read().strip()

        if layout != _BlobCacheLayout.LAYOUT_NAME:
            # Refuse to do anything for the wrong layout. This is especially
            # important since we can be called with `python -m ...` now.
            logger.critical("Invalid blob directory layout %s in %s", layout, blob_dir)
            raise ValueError("Invalid blob directory layout", layout, blob_dir)


        self.blob_dir = blob_dir
        self.target_size = target_size
        self.blob_dir_size = None
        self._finished_callback = when_done
        self.duration = 0.0

        self.__name__ = 'Blob Cache Checker: %s' % (blob_dir,)

    def __acquire_check_lock(self):
        # Returns a lock, or None if we couldn't acquire it.
        blob_dir = self.blob_dir
        lock_path = os.path.join(blob_dir, 'check_size.lock')

        try:
            return zc.lockfile.LockFile(lock_path)
        except zc.lockfile.LockError:
            try:
                time.sleep(1)
                return zc.lockfile.LockFile(lock_path)
            except zc.lockfile.LockError:
                # Someone is already cleaning up, so don't bother
                logger.debug("Another thread is checking the blob cache size.")
                return

    def __size_blob_dir(self, is_cache_dir_name=re.compile(r'\d+$').match):
        # Calculate the sizes of the blobs stored in the blob_dir.
        # Return the total size, and a BTree {atime: [full path to blob file]}

        # TODO: nti.zodb.containers has support for mapping
        # time.time() values into integers for use with the (smaller,
        # faster) IOBTree. Use that if we can prove that we can pop
        # the min atime successfully (that is, while the
        # nti.zodb.containers transformation is lossless and
        # reversible, we need to prove that it also maintains order;
        # I'm not sure it does).
        #
        # Other optimizations: Don't use a list until we get more than one
        # file with a matching atime. And/or use tuples and not lists:
        # tuples aren't tracked by the GC like lists are (after they survive one
        # collection, anyway).

        blob_dir = self.blob_dir
        blob_suffix = ZODB.blob.BLOB_SUFFIX
        files_by_atime = OOBTree.BTree()
        size = 0

        # Use os.walk() instead of os.listdir(); on 3.5+ this is much faster
        # thanks to the use of os.scandir(). When we're on Python 3.5+ *only*
        # we could use os.scandir ourself and maybe save some stat calls?
        for dirpath, dirnames, filenames in os.walk(blob_dir):
            # Walk top-down, only recursing into directories matching the
            # OID components (of which there should be one level)
            dirnames[:] = [d for d in dirnames if is_cache_dir_name(d)]
            # Examine blob files.
            blobfile_paths = [os.path.join(dirpath, f)
                              for f in filenames
                              if f.endswith(blob_suffix)]

            for file_path in blobfile_paths:
                stat = os.stat(file_path)
                size += stat.st_size
                t = stat.st_atime
                if t not in files_by_atime:
                    files_by_atime[t] = []

                # The ZEO version returns a weird version of the path,
                #
                #     os.path.join(dirname, file_name)
                #
                # which it must later re-combine to get an actual path:
                #
                #     os.path.join(blob_dir, file_name)
                #
                # It's not clear why it doesn't return the full path
                # that it already has. Temporary memory savings,
                # perhaps? If so, is that even a concern anymore?
                files_by_atime[t].append(file_path)

        logger.debug("Blob cache size for %s: %s", self.blob_dir, byte_display(size))
        return size, files_by_atime

    @staticmethod
    def remove_blob_at_path(file_path, lock_retries=0):
        """
        Return the size of the blob that was removed, or 0
        if the blob couldn't be removed because it was locked
        or otherwise open (e.g., on Windows).
        """
        try:
            lock = lock_blob(file_path, lock_retries)
        except zc.lockfile.LockError:
            logger.debug("Skipping locked blob %s", file_path)
            return 0  # In use, skip

        try:
            fsize = os.stat(file_path).st_size
            try:
                ZODB.blob.remove_committed(file_path)
            except OSError:
                return 0 # probably open on windows
            else:
                return fsize
        finally:
            lock.close()

    def __shrink_blob_dir(self, current_size, files_by_atime):
        size = current_size
        target_size = self.target_size
        remove = self.remove_blob_at_path

        while size > target_size and files_by_atime:
            for file_path in files_by_atime.pop(files_by_atime.minKey()):
                size -= remove(file_path)
                if size <= target_size:
                    break

        logger.debug("Reduced blob cache size for %s: %s", self.blob_dir, byte_display(size))

    def __call__(self):
        with self:
            logger.info("Checking blob cache size for %s. (target: %s)",
                        self.blob_dir,
                        byte_display(self.target_size))

            check_lock = self.__acquire_check_lock()
            try:
                if check_lock is None:
                    logger.info("Failed to get filesystem clean lock (%s); quitting.",
                                self.blob_dir)
                    return

                self.__run_with_lock()
            finally:
                if check_lock is not None:
                    check_lock.close()
                self._finished_callback(self, check_lock is not None)

    def __run_with_lock(self):
        while 1:
            size, files_by_atime = self.__size_blob_dir()
            self.blob_dir_size = size

            if size <= self.target_size:
                logger.info(
                    'Traversed %s to compute size %s (<= %s); quitting.',
                    self.blob_dir,
                    byte_display(self.blob_dir_size),
                    byte_display(self.target_size)
                )
                break

            self.__shrink_blob_dir(size, files_by_atime)



def main():
    """
    This module can be executed as a main module. It takes two arguments:

    - a cache blob directory; and
    - a target size.

    It will proceed to prune old files from that directory until it reaches the
    target size.

    This is mainly meant to be used with the ``blob_cache_size_check_external``
    option.

    Exits with 0 on success, or a positive value on failure.
    """
    import logging
    import sys
    # Configure basic logging to go to stdout/err so that it can be seen from
    # our parent process when we're launched from a storage.
    logging.basicConfig()

    args = sys.argv[1:]

    cache_dir = args[0]
    target_size = int(args[1])

    checker = _BlobCacheSizeChecker(cache_dir, target_size)
    checker()


if __name__ == '__main__':
    main()



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Pack/Undo implementations.
"""
from __future__ import absolute_import

from ..packundo import HistoryFreePackUndo
from ..packundo import HistoryPreservingPackUndo
from .dialect import OracleDialect


def _oracle_fetchmany(self, cursor): # pylint:disable=unused-argument
    # We can't safely fetch many rows at once without
    # getting 'ProgrammingError: LOB variable no longer valid after subsequent fetch'
    # See https://github.com/zodb/relstorage/issues/30
    return cursor



# Oracle fails to notice that pack_object is now filled and chooses
# the wrong execution plan, completely killing this query on large
# RelStorage databases, unless these hints are included.
_oracle_traverse_graph_optimizer_hint = "/*+ FULL(object_ref) FULL(pack_object) */"


class OracleHistoryPreservingPackUndo(HistoryPreservingPackUndo):

    dialect = OracleDialect()

    _script_choose_pack_transaction = """
        SELECT MAX(tid)
        FROM transaction
        WHERE tid > 0
            AND tid <= %(tid)s
            AND packed = 'N'
        """

    _script_create_temp_pack_visit = None
    _script_create_temp_undo = None
    _script_reset_temp_undo = "DELETE FROM temp_undo"

    _script_find_pack_tid = """
        SELECT MAX(keep_tid)
        FROM pack_object
        """

    _script_transaction_has_data = """
        SELECT DISTINCT tid
        FROM object_state
        WHERE tid = %(tid)s
        """

    _script_delete_empty_transactions_batch = """
        DELETE FROM transaction
        WHERE packed = %(TRUE)s
          AND empty = %(TRUE)s
          AND rownum <= 1000
        """

    # XXX: This is necessary
    # (https://github.com/zodb/relstorage/issues/135), but the HP
    # tests don't fail without it.
    _fetchmany = _oracle_fetchmany

    _traverse_graph_optimizer_hint = _oracle_traverse_graph_optimizer_hint


class OracleHistoryFreePackUndo(HistoryFreePackUndo):

    dialect = OracleDialect()

    _script_choose_pack_transaction = """
        SELECT MAX(tid)
        FROM object_state
        WHERE tid > 0
            AND tid <= %(tid)s
        """

    _script_create_temp_pack_visit = None

    _fetchmany = _oracle_fetchmany
    _traverse_graph_optimizer_hint = _oracle_traverse_graph_optimizer_hint


def _check_limit(c):
    for b in c.mro():
        for k, v in vars(b).items():
            if hasattr(v, 'bind'):
                v = v.bind(c.dialect)
            if 'LIMIT ' in str(v):
                raise TypeError("Forgot to override %s in %s" % (k, c))

_check_limit(OracleHistoryPreservingPackUndo)
_check_limit(OracleHistoryFreePackUndo)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""TransactionControl implementations"""
from __future__ import absolute_import
from __future__ import print_function

import abc

from zope.interface import implementer

from .._compat import ABC

from ._util import noop_when_history_free

from .schema import Schema
from .interfaces import ITransactionControl


class AbstractTransactionControl(ABC):
    """Abstract base class"""

    # pylint:disable=unused-argument

    def __init__(self, connmanager):
        self.connmanager = connmanager

    def commit_phase1(self, store_connection, tid):
        return '-'

    def commit_phase2(self, store_connection, txn, load_connection):
        store_connection.commit()

    def abort(self, store_connection, txn=None):
        """
        Abort the commit. If txn is not None, phase 1 is also aborted.

        The connection is rolled back quietly using :meth:`~IConnectionManager.rollback_quietly`
        and the boolean result of that function is returned.
        """
        return store_connection.rollback_quietly()

    @abc.abstractmethod
    def get_tid(self, cursor):
        "Returns the most recent tid"
        raise NotImplementedError()

    @abc.abstractmethod
    def add_transaction(self, cursor, tid, username, description, extension,
                        packed=False):
        """Add a transaction"""
        raise NotImplementedError()


@implementer(ITransactionControl)
class GenericTransactionControl(AbstractTransactionControl):
    """
    A :class:`ITransactionControl` implementation that works for history-free
    and history-preserving storages that share a common syntax.
    """

    def __init__(self, connmanager, poller, keep_history, Binary):
        super(GenericTransactionControl, self).__init__(connmanager)
        self.poller = poller
        self.keep_history = keep_history
        self.Binary = Binary

    def get_tid(self, cursor):
        return self.poller.get_current_tid(cursor)

    _add_transaction_query = Schema.transaction.insert(
        Schema.transaction.c.tid,
        Schema.transaction.c.packed,
        Schema.transaction.c.username,
        Schema.transaction.c.description,
        Schema.transaction.c.extension
    ).prepared()

    @noop_when_history_free
    def add_transaction(self, cursor, tid, username, description, extension,
                        packed=False):
        binary = self.Binary
        self._add_transaction_query.execute(
            cursor,
            (
                tid, packed, binary(username),
                binary(description), binary(extension)
            )
        )

    _delete_transaction_query = Schema.transaction.delete(
    ).where(
        Schema.transaction.c.tid == Schema.transaction.orderedbindparam()
    )

    @noop_when_history_free
    def delete_transaction(self, cursor, tid):
        self._delete_transaction_query.execute(
            cursor,
            (tid,)
        )



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2016 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
PyMySQL IDBDriver implementations.
"""
from __future__ import absolute_import
from __future__ import print_function

from zope.interface import implementer

from relstorage._util import Lazy
from relstorage.adapters.interfaces import IDBDriver

from . import AbstractMySQLDriver

__all__ = [
    'PyMySQLDriver',
]

# The default conversion looses precision in floating point values,
# truncating them at 5 digits and using an exponent, which also loses
# precision. All other drivers send the full representation. See
# pymysql.converters:escape_float. The version here is based on
# what mysqlclient has.
def escape_float(value, _mapping):
    s = repr(value)
    if s in ('inf', 'nan'):
        raise TypeError("%s can not be used with MySQL" % s)
    if 'e' not in s:
        s += 'e0'
    return s

@implementer(IDBDriver)
class PyMySQLDriver(AbstractMySQLDriver):
    __name__ = 'PyMySQL'
    MODULE_NAME = 'pymysql'
    PRIORITY = 2
    PRIORITY_PYPY = 1
    _GEVENT_CAPABLE = True
    _GEVENT_NEEDS_SOCKET_PATCH = True

    def __init__(self):
        super(PyMySQLDriver, self).__init__()

        pymysql_err = self.driver_module

        # Under PyMySql through at least 0.6.6, closing an already closed
        # connection raises a plain pymysql.err.Error.
        # It can also raise a DatabaseError, and sometimes
        # an IOError doesn't get mapped to a type
        self.close_exceptions += (
            pymysql_err.Error,
            IOError,
            pymysql_err.DatabaseError
        )

        self.disconnected_exceptions += (
            IOError, # This one can escape mapping;
            # This one has only been seen as its subclass,
            # InternalError, as (0, 'Socket receive buffer full'),
            # which should probably be taken as disconnect
            pymysql_err.DatabaseError,
        )

        self._conversions = self.driver_module.converters.conversions.copy()
        self._conversions[float] = escape_float

    def connect(self, *args, **kwargs):
        kwargs['conv'] = self._conversions
        # For bytes and bytearrays, put the "_binary" character set introducer
        # in front. If we haven't set the ``connection_charset`` to binary --- and we can't
        # because that breaks JSON types --- then this is needed for sending state data
        # to avoid warnings about invalid character sequences.
        kwargs['binary_prefix'] = True
        return AbstractMySQLDriver.connect(self, *args, **kwargs)

    @Lazy
    def _server_side_cursor(self):
        # This appears like it might only read one row at a time, which is probably not
        # great. But the usual solution, falling back to fetchmany(),
        # doesn't do any good: fetchmany is implemented in terms of fetchone().
        from pymysql.cursors import SSCursor # pylint:disable=no-name-in-module,import-error

        return SSCursor



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from relstorage.tests import TestCase

from ..query import Query as _BaseQuery
from ..query import CompiledQuery

class MockDialect(object):

    def bind(self, context): # pylint:disable=unused-argument
        return self

class Query(_BaseQuery):

    def compiled(self):
        return self

class TestQuery(TestCase):

    def test_name_discovery(self):
        # If a __name__ isn't assigned when a query is a
        # class property and used as a non-data-descriptor,
        # it finds it.

        class C(object):
            dialect = MockDialect()

            q1 = Query()
            q2 = Query()
            q_over = Query()

        class D(C):

            q3 = Query()
            q_over = Query()

        inst = D()

        # Undo the effects of Python 3.6's __set_name__.
        # pylint:disable=non-str-assignment-to-dunder-name,attribute-defined-outside-init
        D.q1.__name__ = None
        D.q2.__name__ = None
        D.q3.__name__ = None
        C.q_over.__name__ = None
        D.q_over.__name__ = None

        # get them to trigger them to search their name
        getattr(inst, 'q1')
        getattr(inst, 'q2')
        getattr(inst, 'q3')
        getattr(inst, 'q_over')


        self.assertEqual(C.q1.__name__, 'q1')
        self.assertEqual(C.q2.__name__, 'q2')
        self.assertEqual(D.q3.__name__, 'q3')
        self.assertIsNone(C.q_over.__name__)
        self.assertEqual(D.q_over.__name__, 'q_over')


class TestCompiledQuery(TestCase):

    def test_stmt_cache_on_bad_cursor(self):

        unique_execute_stmt = []

        class MockStatement(object):
            class dialect(object):
                class compiler(object):
                    def __init__(self, _):
                        "Does nothing"
                    def compile(self):
                        return 'stmt', ()
                    def can_prepare(self):
                        # We have to prepare if we want to try the cache
                        return True
                    def prepare(self):
                        o = object()
                        unique_execute_stmt.append(o)
                        return "prepare", o, lambda params: params

        executed = []

        class Cursor(object):
            __slots__ = ('__weakref__',)

            @property
            def connection(self):
                return self

            def execute(self, stmt):
                executed.append(stmt)

        cursor = Cursor()

        query = CompiledQuery(MockStatement())
        query.execute(cursor)

        self.assertLength(unique_execute_stmt, 1)
        self.assertLength(executed, 2)
        self.assertEqual(executed, [
            "prepare",
            unique_execute_stmt[0],
        ])

        query.execute(cursor)
        self.assertLength(unique_execute_stmt, 1)
        self.assertLength(executed, 3)
        self.assertEqual(executed, [
            "prepare",
            unique_execute_stmt[0],
            unique_execute_stmt[0],
        ])



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""sqlite3 adapter for RelStorage."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sqlite3
import threading

from zope.interface import implementer

from relstorage._util import consume
from ..interfaces import IOIDAllocator
from ..oidallocator import AbstractRangedOIDAllocator

@implementer(IOIDAllocator)
class Sqlite3OIDAllocator(AbstractRangedOIDAllocator):
    # Even though we use the new_oid table like AbstractTableOIDAllocator
    # does, because we have to take exclusive locks *anyway*, we can
    # ensure that it only ever has a single increasing row.

    def __init__(self, db_path, driver):
        super(Sqlite3OIDAllocator, self).__init__()
        self.db_path = db_path
        self.lock = threading.Lock()
        self.driver = driver
        self._connection = None

    def new_instance(self):
        return self

    def release(self):
        self.close()

    def close(self):
        with self.lock:
            if self._connection:
                self._connection.close()
                self._connection = None

    # This is actually the last OID we returned. We begin allocating beginning with
    # the *next* integer.
    _new_oid_query = 'CREATE TABLE new_oid (zoid INTEGER PRIMARY KEY NOT NULL);'

    def _connect(self):
        if self._connection is None:
            conn = self.driver.connect_to_file(
                self.db_path,
                mmap_size=1024 * 1024 * 10, # try to keep the whole thing in memory.
                override_pragmas={
                    # We can always reconstruct the contents of this file from the database
                    # itself, and speed is utterly critical.
                    'journal_mode': 'off',
                    'synchronous': 'off',
                }
            )
            try:
                consume(conn.execute('SELECT count(*) from new_oid'))
            except sqlite3.OperationalError:
                conn.executescript(
                    self._new_oid_query + """
                INSERT OR REPLACE INTO new_oid
                SELECT MAX(x) FROM (
                    SELECT 0 x
                    UNION ALL
                    SELECT MAX(zoid)
                    FROM new_oid
                )
                """)
            self._connection = conn
        return self._connection

    def _set_min_oid_from_range(self, cursor, n):
        # Recall that the very first write to the database will cause
        # the file to be locked against further writes. So there's some
        # benefit in avoiding writes if they're not needed. Because we
        # keep this in a separate database as well, we can keep the connection
        # in autocommit mode.
        with self.lock:
            # We've left the underlying connection in autocommit mode.
            conn = self._connect()
            rows = conn.execute(
                'SELECT zoid FROM new_oid WHERE zoid < ?',
                (n,)).fetchall()
            if rows:
                # Narf, we need to execute a write transaction.
                consume(conn.execute(
                    'UPDATE new_oid SET zoid = :new WHERE zoid < :new',
                    {'new': n}))

    def new_oids(self, cursor=None):
        return self.new_oids_no_cursor()

    def new_oids_no_cursor(self):
        with self.lock:
            conn = self._connect()
            consume(conn.execute('BEGIN IMMEDIATE TRANSACTION'))
            row, = conn.execute('SELECT zoid FROM new_oid')
            conn.execute('UPDATE new_oid SET zoid = zoid + 1')
            conn.commit()
        return self._oid_range_around(row[0] + 1)

    def reset_oid(self, cursor=None):
        with self.lock:
            self._connect().execute('UPDATE new_oid SET zoid = 0')



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Tests for AST.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from relstorage.tests import TestCase

from .. import ast


class TestFuncs(TestCase):

    def test_as_node_boolean(self):
        node = ast.as_node(True)
        self.assertIsInstance(node, ast.BooleanNode)
        self.assertIs(node.raw, True)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2016 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Oracle IDBDriver implementations.
"""

from __future__ import absolute_import
from __future__ import print_function


from zope.interface import implementer


from ..drivers import AbstractModuleDriver
from ..drivers import implement_db_driver_options
from ..interfaces import IDBDriver
from .dialect import OracleDialect

database_type = 'oracle'

__all__ = [
    'cx_OracleDriver',
]

@implementer(IDBDriver)
class cx_OracleDriver(AbstractModuleDriver):
    __name__ = 'cx_Oracle'
    MODULE_NAME = __name__
    dialect = OracleDialect()

    def __init__(self):
        super(cx_OracleDriver, self).__init__()

        cx_Oracle = self.driver_module

        self.disconnected_exceptions += (cx_Oracle.DatabaseError,)
        self.close_exceptions += (cx_Oracle.DatabaseError,)

        # Extensions
        self.DatabaseError = cx_Oracle.DatabaseError
        self.NUMBER = cx_Oracle.NUMBER
        self.BLOB = cx_Oracle.BLOB
        self.CLOB = cx_Oracle.CLOB
        self.LOB = cx_Oracle.LOB
        self.LONG_BINARY = cx_Oracle.LONG_BINARY
        self.LONG_STRING = cx_Oracle.LONG_STRING
        self.BINARY = cx_Oracle.BINARY
        self.STRING = cx_Oracle.STRING
        self.version = cx_Oracle.version


implement_db_driver_options(
    __name__,
    '.drivers'
)



import os
import platform
import unittest

# ZODB >= 3.9.  The blob directory can be a private cache.
shared_blob_dir_choices = (False, True)

RUNNING_ON_TRAVIS = os.environ.get('TRAVIS')
RUNNING_ON_APPVEYOR = os.environ.get('APPVEYOR')
RUNNING_ON_CI = RUNNING_ON_TRAVIS or RUNNING_ON_APPVEYOR

def _do_not_skip(reason): # pylint:disable=unused-argument
    def dec(f):
        return f
    return dec

if RUNNING_ON_CI:
    skipOnCI = unittest.skip
else:
    skipOnCI = _do_not_skip

if RUNNING_ON_APPVEYOR:
    skipOnAppveyor = unittest.skip
else:
    skipOnAppveyor = _do_not_skip
PYPY = platform.python_implementation() == 'PyPy'

CACHE_SERVERS = None
CACHE_MODULE_NAME = None

if RUNNING_ON_TRAVIS and not PYPY:
    # We expect to have access to a local memcache server
    # on travis. Use it if we can import drivers.
    #
    # This definitely leaks sockets on PyPy, and I don't think it's our fault.
    # pylint:disable=unused-import
    try:
        import pylibmc
        CACHE_SERVERS = ["localhost:11211"]
        CACHE_MODULE_NAME = 'relstorage.pylibmc_wrapper'
    except ImportError:
        try:
            import memcache
            CACHE_SERVERS = ["localhost:11211"]
            CACHE_MODULE_NAME = 'memcache'
        except ImportError:
            pass

USE_SMALL_BLOBS = ((RUNNING_ON_CI # slow here
                    or platform.system() == 'Darwin' # interactive testing
                    or os.environ.get("RS_SMALL_BLOB")) # define
                   and not os.environ.get('RS_LARGE_BLOB'))

# mysqlclient (aka MySQLdb) and possibly other things that
# use libmysqlclient.so will try to connect over the
# default Unix socket that was established when that
# library was compiled if no host is given. But that
# server may not be running, or may not be the one we want
# to use for testing, so explicitly ask it to use TCP
# socket by giving an IP address (using 'localhost' will
# still try to use the socket.) (The TCP port can be bound
# by non-root, but the default Unix socket often requires
# root permissions to open.)
STANDARD_DATABASE_SERVER_HOST = '127.0.0.1'
DEFAULT_DATABASE_SERVER_HOST = os.environ.get('RS_DB_HOST',
                                              STANDARD_DATABASE_SERVER_HOST)


TEST_UNAVAILABLE_DRIVERS = not bool(os.environ.get('RS_SKIP_UNAVAILABLE_DRIVERS'))
if RUNNING_ON_CI:
    TEST_UNAVAILABLE_DRIVERS = False


class MinimalTestLayer(object):

    __bases__ = ()
    __module__ = ''

    def __init__(self, name, module='', base=None):
        self.__name__ = name
        if module:
            self.__module__ = module
        if base:
            self.__bases__ = (base,)

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testSetUp(self):
        pass

    def testTearDown(self):
        pass


GeventLayer = MinimalTestLayer('gevent', __name__)

class _Availability(object):
    """
    Has a boolean value telling whether the driver or database is available,
    and a string explaining why it is/is not.

    Note that this includes checking whether we can connect to the database.
    """



    def __init__(self, factory, drivers, max_priority, use_adapter, db_name,
                 raised_exceptions=(TypeError, AttributeError)):
        from relstorage.adapters.interfaces import DriverNotAvailableError
        self.raised_exceptions = raised_exceptions
        self.driver_name = factory.driver_name
        self.escaped_driver_name = self.driver_name.replace(' ', '').replace('/', '_')

        try:
            self.driver = drivers.select_driver(self.driver_name)
        except DriverNotAvailableError:
            self.driver = None
        else:
            if self.driver.gevent_cooperative() and 'gevent' not in self.driver_name:
                # Pure-python implementations that can do gevent but
                # don't have to still need to have it in their name so we can get them
                # by layer.
                self.escaped_driver_name = 'gevent_' + self.escaped_driver_name

        self._available = self.driver is not None and self.driver.priority <= max_priority

        if not self._available:
            if self.driver is None:
                msg = 'Driver %s is not installed' % (self.driver_name,)
            else:
                msg = 'Driver %s has test priority %d >= max %d' % (
                    self.driver_name, self.driver.priority, max_priority
                )
        else:
            msg = 'Driver %s is installed' % (self.driver_name,)
        self._msg = msg

        if self.driver is not None:
            type(self.driver).STRICT = True

        if self._available:
            # See if we can connect.
            self.__check_db_access(use_adapter, db_name)


    def __str__(self):
        return self._msg

    def __bool__(self):
        return self._available

    __nonzero__ = __bool__

    def __check_db_access_cb(self, _conn, _cursor):
        "Does nothing"

    __check_db_access_cb.transaction_read_only = True

    def __check_db_access(self, use_adapter, db_name):
        # We need to get an adapter to get a connmanager to try to connect.
        from relstorage.options import Options
        options = Options(driver=self.driver_name)

        adapter_maker = use_adapter()
        adapter_maker.driver_name = self.driver_name
        adapter = adapter_maker.make_adapter(options, db_name)
        try:
            adapter.connmanager.open_and_call(self.__check_db_access_cb)
        except self.raised_exceptions:
            # We're called from test_suite(), and zope.testrunner
            # ignores errors at that time, so we need to print it ourself.
            import traceback; traceback.print_exc()
            raise
        except Exception as e:  # pylint:disable=broad-except
            self._available = False
            self._msg = "%s: Failed to connect: %r %s" % (self._msg, type(e), e)


class AbstractTestSuiteBuilder(object):

    __name__ = None # PostgreSQL, MySQL, Oracle
    # Drivers with a priority over this amount won't be part of the
    # test run even if installed.
    MAX_PRIORITY = int(os.environ.get('RS_MAX_TEST_PRIORITY', '100'))

    # Ask the drivers to be in their strictest possible mode.
    STRICT_DRIVER = True

    # The kind of exceptions that are raised, not ignored,
    # when trying to check access to the database. They indicate
    # programming problems.
    RAISED_EXCEPTIONS = (TypeError, AttributeError)

    def __init__(self, driver_options, use_adapter, extra_test_classes=()):
        """
        :param driver_options: The ``IDBDriverOptions``
        :param use_adapter: A mixin class implementing the abstract methods
            defined by ``StorageCreatingMixin``.
        """
        self.database_layer = MinimalTestLayer(self.__name__)
        self.gevent_layer = MinimalTestLayer('gevent_' + self.__name__,
                                             '',
                                             GeventLayer)
        self.gevent_layer.__bases__ += (self.database_layer,)
        self.drivers = driver_options
        self.extra_test_classes = extra_test_classes
        self.base_dbname = os.environ.get('RELSTORAGETEST_DBNAME', 'relstoragetest')
        self.db_names = {
            'data': self.base_dbname,
            '1': self.base_dbname,
            '2': self.base_dbname + '2',
            'dest': self.base_dbname + '2',
        }

        self.use_adapter = use_adapter
        use_adapter.base_dbname = self.base_dbname
        self.large_blob_size = self._compute_large_blob_size(USE_SMALL_BLOBS)

    def _compute_large_blob_size(self, use_small_blobs):
        raise NotImplementedError

    def test_suite(self):
        from .reltestbase import AbstractIDBDriverTest
        from .reltestbase import AbstractIDBOptionsTest

        suite = unittest.TestSuite()
        suite.addTest(unittest.makeSuite(type(
            self.__name__ + 'DBOptionsTest',
            (AbstractIDBOptionsTest,),
            {'db_options': self.drivers}
        )))
        for factory in self.drivers.known_driver_factories():
            available = _Availability(
                factory, self.drivers, self.MAX_PRIORITY,
                self.use_adapter,
                self.db_names['data'],
                self.RAISED_EXCEPTIONS
            )
            # Checking the driver is just a unit test, it doesn't connect or
            # need a layer
            suite.addTest(unittest.makeSuite(
                self.__skipping_if_not_available(
                    type(
                        self.__name__ + 'DBDriverTest_' + available.escaped_driver_name,
                        (AbstractIDBDriverTest,),
                        {'driver': available.driver}
                    ),
                    available.driver is not None)))

            # On CI, we don't even add tests for unavailable drivers to the
            # list of tests; this makes the output much shorter and easier to read,
            # but it does make zope-testrunner's discovery options less useful.
            if not available and not TEST_UNAVAILABLE_DRIVERS:
                continue

            # We put the various drivers into a zope.testrunner layer
            # for ease of selection by name, e.g.,
            # zope-testrunner --layer PG8000Driver
            driver_suite = unittest.TestSuite()
            layer_name = available.escaped_driver_name

            base = self.database_layer
            if available.driver and available.driver.gevent_cooperative():
                base = self.gevent_layer

            driver_suite.layer = MinimalTestLayer(layer_name,
                                                  base.__name__ + '.drivers',
                                                  base)
            self._add_driver_to_suite(driver_suite, layer_name, available)
            suite.addTest(driver_suite)
        return suite

    def _default_make_check_class(self, bases, name, klass_dict=None):
        base_klass_dict = self._make_base_klass_dict()
        if klass_dict:
            base_klass_dict.update(klass_dict)

        klass = type(
            name,
            (self.use_adapter,) + bases,
            base_klass_dict
        )

        return klass

    _default_make_test_class = _default_make_check_class

    def _make_base_klass_dict(self):
        return {}

    def __make_test_class(self, base, extra_bases, maker_base_name, maker_default, klass_dict=None):
        name = self.__name__ + base.__name__
        maker = getattr(
            self,
            maker_base_name + base.__name__,
            maker_default
        )
        __traceback_info__ = maker, base
        klass = maker((base,) + extra_bases, name, klass_dict)
        klass.__module__ = self.__module__
        klass.__name__ = name
        return klass

    def _make_check_classes(self):
        # The classes that inherit from ZODB tests and use 'check' instead of 'test_'

        # This class  is sadly not super() cooperative, so we must
        # try to explicitly put it last in the MRO.
        from ZODB.tests.util import TestCase as ZODBTestCase
        from .hftestbase import HistoryFreeFromFileStorage
        from .hftestbase import HistoryFreeToFileStorage
        from .hftestbase import HistoryFreeRelStorageTests

        from .hptestbase import HistoryPreservingFromFileStorage
        from .hptestbase import HistoryPreservingToFileStorage
        from .hptestbase import HistoryPreservingRelStorageTests

        classes = []

        for _, bases in (
                ('HF', (HistoryFreeFromFileStorage,
                        HistoryFreeToFileStorage,
                        HistoryFreeRelStorageTests)),
                ('HP', (HistoryPreservingFromFileStorage,
                        HistoryPreservingToFileStorage,
                        HistoryPreservingRelStorageTests))
        ):
            for base in bases:
                classes.append(self.__make_test_class(
                    base,
                    (ZODBTestCase,),
                    '_make_check_class_',
                    self._default_make_check_class,
                ))

        return classes

    def _make_test_classes(self):
        # Classes that inherit from RelStorageTestBase and not the ZODB
        # tests. These use the standard convention of 'test_' methods instead
        # of 'check' and don't have ridiculous class hierarchies.
        from .packundo import HistoryFreeTestPack
        from .packundo import HistoryPreservingTestPack

        classes = []


        for base in (
                HistoryFreeTestPack,
                HistoryPreservingTestPack,
        ):
            classes.append(self.__make_test_class(
                base,
                (),
                '_make_test_class_',
                self._default_make_test_class,
            ))

        return classes

    def _make_zodbconvert_classes(self):
        from .reltestbase import AbstractRSDestZodbConvertTests
        from .reltestbase import AbstractRSSrcZodbConvertTests

        classes = []
        for base in (AbstractRSSrcZodbConvertTests, AbstractRSDestZodbConvertTests):
            klass = type(
                self.__name__ + base.__name__[8:],
                (self.use_adapter, base),
                {}
            )
            klass.__module__ = self.__module__
            classes.append(klass)
        return classes

    def __skipping_if_not_available(self, klass, availability):
        klass.__module__ = self.__module__
        klass = unittest.skipUnless(
            availability,
            str(availability))(klass)
        return klass

    def _new_class_for_driver(self, base, driver_available):
        klass = type(
            base.__name__ + '_' + driver_available.escaped_driver_name,
            (base,),
            {'driver_name': driver_available.driver_name}
        )

        return self.__skipping_if_not_available(klass, driver_available)

    def _add_driver_to_suite(self, suite, layer_prefix, driver_available):
        for klass in self._make_check_classes():
            klass = self._new_class_for_driver(klass, driver_available)
            suite.addTest(unittest.makeSuite(klass, "check"))

        for klass in self._make_test_classes():
            klass = self._new_class_for_driver(klass, driver_available)
            suite.addTest(unittest.makeSuite(klass))

        for klass in self._make_zodbconvert_classes():
            suite.addTest(unittest.makeSuite(
                self._new_class_for_driver(klass,
                                           driver_available)))

        for klass in self.extra_test_classes:
            suite.addTest(unittest.makeSuite(
                self._new_class_for_driver(klass,
                                           driver_available)))

        from relstorage.tests.blob.testblob import storage_reusable_suite
        from relstorage.options import Options
        from relstorage.storage import RelStorage

        for shared_blob_dir in shared_blob_dir_choices:
            for keep_history in (False, True):
                # TODO: Make any of the tests that are needing this
                # subclass StorageCreatingMixin so we unify where
                # that's handled.
                history_layer = MinimalTestLayer(
                    '%s%s' % (
                        'Shared' if shared_blob_dir else 'Unshared',
                        'HistoryPreserving' if keep_history else 'HistoryFree',
                    ),
                    suite.layer.__module__ + '.' + suite.layer.__name__ + '.blobs',
                    suite.layer
                )
                layer_name = history_layer.__name__
                def create_storage(name, blob_dir,
                                   shared_blob_dir=shared_blob_dir,
                                   keep_history=keep_history, **kw):
                    if not driver_available:
                        raise unittest.SkipTest(str(driver_available))
                    assert 'driver' not in kw
                    kw['driver'] = driver_available.driver_name
                    db = self.db_names[name]
                    if not keep_history:
                        db += '_hf'

                    options = Options(
                        keep_history=keep_history,
                        shared_blob_dir=shared_blob_dir,
                        blob_dir=os.path.abspath(blob_dir),
                        **kw)

                    adapter_maker = self.use_adapter()
                    adapter_maker.driver_name = driver_available.driver_name
                    adapter = adapter_maker.make_adapter(options, db)
                    __traceback_info__ = adapter, options
                    storage = RelStorage(adapter, name=name, options=options)
                    storage.zap_all()
                    return storage

                prefix = '%s_%s' % (
                    layer_prefix,
                    layer_name
                )

                # If the blob directory is a cache, don't test packing,
                # since packing can not remove blobs from all caches.
                test_packing = shared_blob_dir

                blob_suite = storage_reusable_suite(
                    prefix, create_storage,
                    keep_history=keep_history,
                    test_blob_storage_recovery=True,
                    test_packing=test_packing,
                    test_undo=keep_history,
                    test_blob_cache=(not shared_blob_dir),
                    # PostgreSQL blob chunks are max 2GB in size
                    large_blob_size=(not shared_blob_dir) and (self.large_blob_size) + 100,
                    storage_is_available=driver_available
                )

                blob_suite.layer.__bases__ = blob_suite.layer.__bases__ + (history_layer,)
                blob_suite.layer.__module__ = "%s.%s" % (
                    history_layer.__module__,
                    history_layer.__name__)
                suite.addTest(blob_suite)

        return suite



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2008, 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Implementation of store-related methods.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from .._compat import metricmethod
from .._compat import metricmethod_sampled
from .util import writable_storage_method
from .util import phase_dependent_aborts_early

logger = __import__('logging').getLogger(__name__)

class Storer(object):

    __slots__ = ()

    @phase_dependent_aborts_early
    @writable_storage_method
    @metricmethod_sampled
    def store(self, tpc_phase, oid, previous_tid, data, version, transaction):
        # Called by Connection.commit(), after tpc_begin has been called.
        assert not version, "Versions aren't supported"

        # If we get here and we're read-only, our phase will report that.
        tpc_phase.store(oid, previous_tid, data, transaction)

    @phase_dependent_aborts_early
    @writable_storage_method
    def restore(self, tpc_phase, oid, serial, data, version, prev_txn, transaction):
        # Like store(), but used for importing transactions.  See the
        # comments in FileStorage.restore().  The prev_txn optimization
        # is not used.
        # pylint:disable=unused-argument
        assert not version, "Versions aren't supported"
        # If we get here and we're read-only, our phase will report that.

        tpc_phase.restore(oid, serial, data, prev_txn, transaction)

    @phase_dependent_aborts_early
    @writable_storage_method
    def deleteObject(self, tpc_phase, oid, oldserial, transaction):
        # This method is only expected to be called from zc.zodbdgc
        # currently.
        return tpc_phase.deleteObject(oid, oldserial, transaction)


class BlobStorer(object):

    __slots__ = (
    )

    @phase_dependent_aborts_early
    @writable_storage_method
    @metricmethod
    def storeBlob(self, tpc_phase, oid, serial, data, blobfilename, version, txn):
        """
        Stores data that has a BLOB attached.

        The blobfilename argument names a file containing blob data.
        The storage will take ownership of the file and will rename it
        (or copy and remove it) immediately, or at transaction-commit
        time.  The file must not be open.

        Returns nothing.
        """
        assert not version
        # We used to flush the batcher here, for some reason.
        store_func = tpc_phase.store
        cursor = tpc_phase.shared_state.store_connection.cursor
        blobhelper = tpc_phase.shared_state.blobhelper
        blobhelper.storeBlob(cursor, store_func,
                             oid, serial, data, blobfilename, version, txn)

    @phase_dependent_aborts_early
    @writable_storage_method
    def restoreBlob(self, tpc_phase, oid, serial, data, blobfilename, prev_txn, txn):
        """
        Write blob data already committed in a separate database

        See the restore and storeBlob methods.
        """
        tpc_phase.restoreBlob(oid, serial, data, blobfilename, prev_txn, txn)



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
MySQL adapter for RelStorage.
"""
from __future__ import absolute_import
from __future__ import print_function

import logging

from ..connmanager import AbstractConnectionManager

log = logging.getLogger(__name__)

class MySQLdbConnectionManager(AbstractConnectionManager):

    # https://dev.mysql.com/doc/refman/5.7/en/innodb-transaction-isolation-levels.html

    # Each statement gets its own snapshot. Persistent locking is minimized to
    # the exact index records found (assuming they were found with unique queries.
    # So only query on primary key or unique indexes).
    isolation_read_committed = "ISOLATION LEVEL READ COMMITTED"
    # (DEFAULT) A snapshot is taken when the first statement is
    # executed. All reads are from that snapshot. You can use 'START
    # TRANSACTION WITH CONSISTENT SNAPSHOT' to take a snapshot immediately.
    # If read-only, supposed to barely use locks.
    isolation_repeatable_read = "ISOLATION LEVEL REPEATABLE READ"
    # READ ONLY was new in 5.6
    isolation_repeatable_read_ro = isolation_repeatable_read + ' , READ ONLY'
    # Serializable takes a snapshot, but also implicitly adds 'FOR
    # SHARE' to the end of every SELECT statement (even in READ ONLY
    # mode) --- but only for specific queries like 'SELECT * FROM
    # object_state WHERE zoid = X'; range queries, OTOH, that return
    # the same rows, are unaffected. Therefore, it fails to achieve
    # concurrency with writes.
    isolation_serializable = 'ISOLATION LEVEL SERIALIZABLE'

    def __init__(self, driver, params, options):
        self._params = params.copy()
        self._db_connect = driver.connect
        self._db_driver = driver
        super(MySQLdbConnectionManager, self).__init__(options, driver)

        self.isolation_load = self.isolation_repeatable_read_ro
        self.isolation_store = self.isolation_read_committed

    def _alter_params(self, replica):
        """Alter the connection parameters to use the specified replica.

        The replica parameter is a string specifying either host or host:port.
        """
        params = self._params.copy()
        if ':' in replica:
            host, port = replica.split(':')
            params['host'] = host
            params['port'] = int(port)
        else:
            params['host'] = replica
        return params

    def open(self,
             isolation=None,
             deferrable=False,
             read_only=False,
             replica_selector=None,
             application_name=None,
             **kwargs):
        """Open a database connection and return (conn, cursor)."""
        # pylint:disable=arguments-differ,unused-argument
        if replica_selector is None:
            replica_selector = self.replica_selector

        if replica_selector is not None:
            replica = replica_selector.current()
            params = self._alter_params(replica)
        else:
            replica = None
            params = self._params

        if isolation is None:
            isolation = self.isolation_load
        if read_only and 'READ ONLY' not in isolation:
            isolation += ' , READ ONLY'

        while True:
            __traceback_info__ = {
                k: v if k != 'passwd' else '<*****>'
                for k, v in params.items()
            }

            try:
                conn = self._db_connect(**params)
                cursor = self.cursor_for_connection(conn)

                self._db_driver.set_autocommit(conn, True)
                # Transaction isolation cannot be changed inside a
                # transaction. 'SET SESSION' changes it for all
                # upcoming transactions.
                stmt = "SET SESSION TRANSACTION %s" % isolation
                __traceback_info__ = stmt
                cursor.execute(stmt)
                self._db_driver.set_autocommit(conn, False)
                conn.replica = replica
                conn.readonly = read_only
                return conn, cursor
            except self.driver.use_replica_exceptions as e:
                if replica is not None:
                    next_replica = replica_selector.next()
                    if next_replica is not None:
                        log.warning("Unable to connect to replica %s: %s, "
                                    "now trying %s", replica, e, next_replica)
                        replica = next_replica
                        params = self._alter_params(replica)
                        continue
                log.warning("Unable to connect: %s", e)
                raise

    def _do_open_for_load(self):
        return self.open(
            isolation=self.isolation_load,
            read_only=True,
            replica_selector=self.ro_replica_selector)

    def rollback_store_quietly(self, conn, cur):
        # MySQL leaks state in temporary tables across
        # transactions, so if we don't drop the connection,
        # we need to clean up temp tables (if they exist!)
        if self.rollback_quietly(conn, cur):
            try:
                cur.execute('CALL clean_temp_state(true)')
                cur.fetchall()
            except self.driver.driver_module.Error:
                log.debug("Failed to clean temp state; maybe not exists yet?")
